# pylint: disable=import-outside-toplevel
from __future__ import annotations

import logging
import tempfile
import zipfile
from collections.abc import Callable
from functools import cached_property
from pathlib import Path

from databricks.labs.blueprint.entrypoint import is_in_debug

from databricks.labs.ucx.framework.utils import run_command
from databricks.labs.ucx.source_code.graph import (
    LibraryResolver,
    DependencyProblem,
)
from databricks.labs.ucx.source_code.path_lookup import PathLookup
from databricks.labs.ucx.source_code.known import KnownList

logger = logging.getLogger(__name__)


class PythonLibraryResolver(LibraryResolver):
    """Resolve python libraries by registering and installing Python libraries."""

    def __init__(
        self,
        allow_list: KnownList,
        runner: Callable[[str | list[str]], tuple[int, str, str]] = run_command,
    ) -> None:
        self._allow_list = allow_list
        self._runner = runner

    def register_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        """We delegate to pip to install the library and augment the path look-up to resolve the library at import.
        This gives us the flexibility to install any library that is not in the allow-list, and we don't have to
        bother about parsing cross-version dependencies in our code."""
        if len(libraries) == 0:
            return []
        if len(libraries) == 1:  # Multiple libraries might be installation flags
            compatibility = self._allow_list.distribution_compatibility(libraries[0])
            if compatibility.known:
                return compatibility.problems
        return self._install_library(path_lookup, *libraries)

    @cached_property
    def _temporary_virtual_environment(self) -> Path:
        # TODO: for `databricks labs ucx lint-local-code`, detect if we already have a virtual environment
        # and use that one. See Databricks CLI code for the labs command to see how to detect the virtual
        # environment. If we don't have a virtual environment, create a temporary one.
        # simulate notebook-scoped virtual environment
        lib_install_folder = tempfile.mkdtemp(prefix='ucx-')
        return Path(lib_install_folder).resolve()

    def _install_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        """Pip install library and augment path look-up to resolve the library at import"""
        path_lookup.append_path(self._temporary_virtual_environment)
        resolved_libraries = self._resolve_libraries(path_lookup, *libraries)
        if libraries[0].endswith(".egg"):
            return self._install_egg(*resolved_libraries)
        return self._install_pip(*resolved_libraries)

    def _resolve_libraries(self, path_lookup: PathLookup, *libraries: str) -> list[str]:
        """Resolve pip installs as library (pip install databricks-labs-ucx) or as path (pip install ../../ucx.whl).

        A library is defined as library, i.e. *not* as path, when the library:
        1. Is not found in the path lookup.
        2. Is not located in the temporary virtual environment where this resolver install libraries. If this is the
           case, it signals the library is installed for a second time.

        Otherwise, we consider the library to be a path.

        Note: The above works given our design choice to *only* build dependency graphs when all dependencies are found.
        """
        libs = []
        for library in libraries:
            maybe_library = path_lookup.resolve(Path(library))
            if maybe_library is None or maybe_library.is_relative_to(self._temporary_virtual_environment):
                libs.append(library)
            else:
                libs.append(maybe_library.as_posix())
        return libs

    def _install_pip(self, *libraries: str) -> list[DependencyProblem]:
        args = [
            "pip",
            "--disable-pip-version-check",
            "install",
            *libraries,
            "-t",
            str(self._temporary_virtual_environment),
            "--upgrade",  # Upgrades when library already installed
        ]
        return_code, stdout, stderr = self._runner(args)
        if stdout or stderr:
            logger.debug(f"pip output:\n{stdout}\n{stderr}")
        if return_code != 0:
            command = " ".join(args)
            problem = DependencyProblem("library-install-failed", f"'{command}' failed with '{stderr}'")
            return [problem]
        return []

    def _install_egg(self, *libraries: str) -> list[DependencyProblem]:
        """Install eggs using easy_install.

        Sources:
            See easy_install in setuptools.command.easy_install
        """
        if len(libraries) > 1:
            problems = []
            for library in libraries:
                problems.extend(self._install_egg(library))
            return problems
        library = libraries[0]

        verbosity = "--verbose" if is_in_debug() else "--quiet"
        easy_install_arguments = [
            "easy_install",
            verbosity,
            "--always-unzip",
            "--install-dir",
            self._temporary_virtual_environment.as_posix(),
            library,
        ]
        setup = self._get_setup()
        if callable(setup):
            try:
                setup(script_args=easy_install_arguments)
                return []
            except (SystemExit, ImportError, ValueError) as e:
                logger.warning(f"Failed to install {library} with (setuptools|distutils).setup, unzipping instead: {e}")
        library_folder = self._temporary_virtual_environment / Path(library).name
        library_folder.mkdir(parents=True, exist_ok=True)
        try:
            # Not a "real" install, though, the best effort to still lint eggs without dependencies
            with zipfile.ZipFile(library, "r") as zip_ref:
                zip_ref.extractall(library_folder)
        except (zipfile.BadZipfile, FileNotFoundError) as e:
            problem = DependencyProblem("library-install-failed", f"Failed to install {library}: {e}")
            return [problem]
        return []

    @staticmethod
    def _get_setup() -> Callable | None:
        try:
            # Mypy can't analyze setuptools due to missing type hints
            from setuptools import setup  # type: ignore
        except (ImportError, AssertionError):
            try:
                # pylint: disable-next=deprecated-module
                from distutils.core import setup  # type: ignore
            except ImportError:
                logger.warning("Could not import setup.")
                setup = None
        return setup

    def __str__(self) -> str:
        return f"PythonLibraryResolver(allow_list={self._allow_list})"



from __future__ import annotations

import dataclasses
import logging
import re
import shlex
from abc import ABC, abstractmethod
from ast import parse as parse_python
from enum import Enum
from pathlib import Path

from astroid import NodeNG  # type: ignore
from sqlglot import parse as parse_sql, ParseError as SQLParseError

from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.source_code.base import NOTEBOOK_HEADER
from databricks.labs.ucx.source_code.graph import (
    DependencyGraph,
    DependencyProblem,
    DependencyGraphContext,
    InheritedContext,
)
from databricks.labs.ucx.source_code.python.python_analyzer import PythonCodeAnalyzer
from databricks.labs.ucx.source_code.notebooks.magic import MagicNode, MagicCommand, magic_command_factory

# use a specific logger for sqlglot warnings so we can disable them selectively
sqlglot_logger = logging.getLogger(f"{__name__}.sqlglot")
logger = logging.getLogger(__name__)

CELL_SEPARATOR = "COMMAND ----------"
MAGIC_PREFIX = 'MAGIC'
LANGUAGE_PREFIX = '%'
LANGUAGE_PI = 'LANGUAGE'
COMMENT_PI = 'COMMENT'


class Cell(ABC):

    def __init__(self, source: str, original_offset: int):
        # The header line prevents a cell from having a zero offset.
        if original_offset < 1:
            raise ValueError("Cells must have a positive offset within the original file.")
        self._original_offset = original_offset
        self._original_code = source
        self._migrated_code = source

    @property
    def original_offset(self) -> int:
        """Line offset of this cell within the original notebook file.

        Example: if the line offset is 5, then line 4 within this cell corresponds to line 9 of the original file.
        """
        return self._original_offset

    @property
    def original_code(self) -> str:
        return self._original_code

    @property
    def migrated_code(self) -> str:
        return self._migrated_code

    @migrated_code.setter
    def migrated_code(self, value: str) -> None:
        self._migrated_code = value

    @property
    @abstractmethod
    def language(self) -> CellLanguage: ...

    @abstractmethod
    def is_runnable(self) -> bool: ...

    def build_dependency_graph(self, _: DependencyGraph) -> list[DependencyProblem]:
        """Check for any problems with dependencies of this cell.

        Returns:
            A list of found dependency problems; position information for problems is relative to the enclosing notebook.
        """
        return []

    def __repr__(self):
        return f"{self.language.name}: {self._original_code[:20]}"

    def build_inherited_context(self, _graph: DependencyGraph, _child_path: Path) -> InheritedContext:
        return InheritedContext(None, False, [])


class PythonCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.PYTHON

    def is_runnable(self) -> bool:
        try:
            tree = parse_python(self._original_code)
            return tree is not None
        except SyntaxError:
            return True

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        context = parent.new_dependency_graph_context()
        analyzer = PythonCodeAnalyzer(context, self._original_code)
        # Position information for the Python code is within the code and needs to be mapped to the location within
        # the parent nodebook.
        problems = []
        for problem in analyzer.build_graph():
            problem = dataclasses.replace(
                problem,
                start_line=self.original_offset + problem.start_line,
                end_line=self.original_offset + problem.end_line,
            )
            problems.append(problem)
        return problems

    def build_inherited_context(self, graph: DependencyGraph, child_path: Path) -> InheritedContext:
        context = graph.new_dependency_graph_context()
        analyzer = PythonCodeAnalyzer(context, self._original_code)
        return analyzer.build_inherited_context(child_path)


class RCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.R

    def is_runnable(self) -> bool:
        return True


class ScalaCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.SCALA

    def is_runnable(self) -> bool:
        return True


class SQLCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.SQL

    def is_runnable(self) -> bool:
        try:
            statements = parse_sql(self._original_code)
            return len(statements) > 0
        except SQLParseError as e:
            sqlglot_logger.warning(f"Failed to parse SQL using 'sqlglot': {self._original_code}", exc_info=e)
            return True


class MarkdownCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.MARKDOWN

    def is_runnable(self) -> bool:
        return True  # TODO


class RunCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.RUN

    def is_runnable(self) -> bool:
        return True  # TODO

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        path, idx, line = self._read_notebook_path()
        if path is not None:
            start_line = self._original_offset + idx
            problems = []
            for problem in parent.register_notebook(path, True):
                problem = dataclasses.replace(
                    problem,
                    start_line=start_line,
                    start_col=0,
                    end_line=start_line,
                    end_col=len(line),
                )
                problems.append(problem)
            return problems
        start_line = self._original_offset
        problem = DependencyProblem(
            'invalid-run-cell',
            "Missing notebook path in %run command",
            start_line=start_line,
            start_col=0,
            end_line=start_line,
            end_col=len(self._original_code),
        )
        return [problem]

    def maybe_notebook_path(self) -> Path | None:
        path, _, _ = self._read_notebook_path()
        return path

    def _read_notebook_path(self) -> tuple[Path | None, int, str]:
        command = f'{LANGUAGE_PREFIX}{self.language.magic_name}'
        lines = self._original_code.split('\n')
        for idx, line in enumerate(lines):
            start = line.find(command)
            if start >= 0:
                path = line[start + len(command) :]
                path = path.strip().strip("'").strip('"')
                if len(path) == 0:
                    continue
                return Path(path), idx, line
        return None, 0, ""


class ShellCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.SHELL

    def is_runnable(self) -> bool:
        return True  # TODO


class PipCell(Cell):

    @property
    def language(self) -> CellLanguage:
        return CellLanguage.PIP

    def is_runnable(self) -> bool:
        return True  # TODO

    def build_dependency_graph(self, graph: DependencyGraph) -> list[DependencyProblem]:
        node = MagicNode(0, 1, None, end_lineno=0, end_col_offset=len(self.original_code))
        return PipCommand(node, self.original_code).build_dependency_graph(graph)


class CellLanguage(Enum):
    # long magic_names must come first to avoid shorter ones being matched
    PYTHON = Language.PYTHON, 'python', '#', True, PythonCell
    SCALA = Language.SCALA, 'scala', '//', True, ScalaCell
    SQL = Language.SQL, 'sql', '--', True, SQLCell
    RUN = None, 'run', '', False, RunCell
    PIP = None, 'pip', '', False, PipCell
    SHELL = None, 'sh', '', False, ShellCell
    # see https://spec.commonmark.org/0.31.2/#html-comment
    MARKDOWN = None, 'md', "<!--->", False, MarkdownCell
    R = Language.R, 'r', '#', True, RCell

    def __init__(self, *args):
        super().__init__()
        self._language = args[0]
        self._magic_name = args[1]
        self._comment_prefix = args[2]
        # PI stands for Processing Instruction
        self._requires_isolated_pi = args[3]
        self._new_cell = args[4]

    @property
    def file_magic_header(self) -> str:
        return f"{self._comment_prefix} {NOTEBOOK_HEADER}"

    @property
    def language(self) -> Language:
        return self._language

    @property
    def magic_name(self) -> str:
        return self._magic_name

    @property
    def comment_prefix(self) -> str:
        return self._comment_prefix

    @property
    def requires_isolated_pi(self) -> str:
        return self._requires_isolated_pi

    @classmethod
    def of_language(cls, language: Language) -> CellLanguage:
        for cell_language in CellLanguage:
            if cell_language.language == language:
                return cell_language
        msg = f"Unsupported cell language: {language}"
        raise ValueError(msg)

    @classmethod
    def of_magic_name(cls, magic_name: str) -> CellLanguage | None:
        return next((cl for cl in CellLanguage if magic_name.startswith(cl.magic_name)), None)

    def read_cell_language(self, lines: list[str]) -> CellLanguage | None:
        magic_prefix = f'{self.comment_prefix} {MAGIC_PREFIX} '
        magic_language_prefix = f'{magic_prefix}{LANGUAGE_PREFIX}'
        for line in lines:
            # if we find a non-comment then we're done
            if not line.startswith(self.comment_prefix):
                return None
            # if it's not a magic comment, skip it
            if not line.startswith(magic_prefix):
                continue
            if line.startswith(magic_language_prefix):
                line = line[len(magic_language_prefix) :]
                return CellLanguage.of_magic_name(line.strip())
            return None
        return None

    def new_cell(self, source: str, original_offset: int) -> Cell:
        return self._new_cell(source, original_offset)

    def extract_cells(self, source: str) -> list[Cell]:
        lines = source.split('\n')
        if not lines[0].startswith(self.file_magic_header):
            raise ValueError("Not a Databricks notebook source!")

        def make_cell(cell_lines: list[str], start: int) -> Cell:
            # trim leading blank lines
            while len(cell_lines) > 0 and len(cell_lines[0]) == 0:
                cell_lines.pop(0)
                start += 1

            # trim trailing blank lines
            while len(cell_lines) > 0 and len(cell_lines[-1]) == 0:
                cell_lines.pop(-1)
            cell_language = self.read_cell_language(cell_lines)
            if cell_language is None:
                cell_language = self
            else:
                cell_lines = self._remove_magic_wrapper(cell_lines, cell_language)

            cell_source = '\n'.join(cell_lines)
            cell = cell_language.new_cell(cell_source, start)
            return cell

        cells = []
        cell_lines: list[str] = []
        separator = f"{self.comment_prefix} {CELL_SEPARATOR}"

        # Start iterating from the 2nd line (after the header) to split the cells.
        next_cell_pos = 1
        for i in range(1, len(lines)):
            line = lines[i].strip()
            if line.startswith(separator):
                cell = make_cell(cell_lines, next_cell_pos)
                cells.append(cell)
                cell_lines = []
                # i points to the cell separator, the next cell begins 1 line later.
                next_cell_pos = i + 1
            else:
                cell_lines.append(lines[i])

        if len(cell_lines) > 0:
            cell = make_cell(cell_lines, next_cell_pos)
            cells.append(cell)

        return cells

    def _process_line(
        self,
        line: str,
        prefix: str,
        lang_prefix: str,
        cell_language: CellLanguage,
    ) -> list[str]:
        if line.startswith(prefix):
            line = line[len(prefix) :]
            if cell_language.requires_isolated_pi and line.startswith(lang_prefix):
                new_line = line[len(lang_prefix) :].strip()
                if new_line:
                    return [f"{cell_language.comment_prefix} {LANGUAGE_PI}", new_line.strip()]
                return [f"{cell_language.comment_prefix} {LANGUAGE_PI}"]
            return [line]
        if line.startswith(self.comment_prefix):
            return [f"{cell_language.comment_prefix} {COMMENT_PI}{line}"]
        return [line]

    def _remove_magic_wrapper(self, lines: list[str], cell_language: CellLanguage) -> list[str]:
        prefix = f"{self.comment_prefix} {MAGIC_PREFIX} "
        lang_prefix = f"{LANGUAGE_PREFIX}{cell_language.magic_name}"
        new_lines = []
        for line in lines:
            new_lines.extend(self._process_line(line, prefix, lang_prefix, cell_language))
        return new_lines

    def wrap_with_magic(self, code: str, cell_language: CellLanguage) -> str:
        language_pi_prefix = f"{cell_language.comment_prefix} {LANGUAGE_PI}"
        comment_pi_prefix = f"{cell_language.comment_prefix} {COMMENT_PI}"
        comment_pi_prefix_len = len(comment_pi_prefix)
        lines = code.split('\n')
        for i, line in enumerate(lines):
            if line.startswith(language_pi_prefix):
                line = f"{self.comment_prefix} {MAGIC_PREFIX} {LANGUAGE_PREFIX}{cell_language.magic_name}"
                lines[i] = line
                continue
            if line.startswith(comment_pi_prefix):
                lines[i] = line[comment_pi_prefix_len:]
                continue
            line = f"{self.comment_prefix} {MAGIC_PREFIX} {line}"
            lines[i] = line
        if code.endswith('./'):
            lines.append('\n')
        return "\n".join(lines)


class RunCommand(MagicCommand):

    @staticmethod
    @magic_command_factory
    def factory(command: str, node: NodeNG) -> MagicCommand | None:
        if command.startswith("%run"):
            return RunCommand(node, command)
        return None

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        path = self.notebook_path
        if path is not None:
            problems = parent.register_notebook(path, True)
            return [problem.from_node(problem.code, problem.message, self._node) for problem in problems]
        problem = DependencyProblem.from_node('invalid-run-cell', "Missing notebook path in %run command", self._node)
        return [problem]

    @property
    def notebook_path(self) -> Path | None:
        start = self._code.find(' ')
        if start < 0:
            return None
        path = self._code[start + 1 :].strip().strip('"').strip("'")
        return Path(path)

    def build_inherited_context(self, context: DependencyGraphContext, child_path: Path) -> InheritedContext:
        path = self.notebook_path
        if path is None:
            problem = DependencyProblem('missing-notebook-path', "Missing notebook path in %run command")
            return InheritedContext(None, False, [problem])
        maybe = context.resolver.resolve_notebook(context.path_lookup, path, inherit_context=True)
        if not maybe.dependency:
            problem = DependencyProblem('missing-notebook', f"Could not locate notebook {path}")
            return InheritedContext(None, False, [problem])
        child_path = maybe.dependency.path
        absolute_path = context.path_lookup.resolve(path)
        absolute_child = context.path_lookup.resolve(child_path)
        if absolute_path == absolute_child:
            return InheritedContext(None, True, [])
        container = maybe.dependency.load(context.path_lookup)
        if not container:
            problem = DependencyProblem('corrupt-notebook', f"Could not load notebook {path}")
            return InheritedContext(None, False, [problem])
        return container.build_inherited_context(context.parent, child_path)


class PipCommand(MagicCommand):

    @staticmethod
    @magic_command_factory
    def factory(command: str, node: NodeNG) -> MagicCommand | None:
        if command.startswith("%pip") or command.startswith("!pip"):
            return PipCommand(node, command)
        return None

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        argv = self._split(self._code)
        if len(argv) == 1:
            return [DependencyProblem.from_node("library-install-failed", "Missing command after 'pip'", self._node)]
        if argv[1] != "install":
            return [
                DependencyProblem.from_node(
                    "library-install-failed", f"Unsupported 'pip' command: {argv[1]}", self._node
                )
            ]
        if len(argv) == 2:
            return [
                DependencyProblem.from_node(
                    "library-install-failed", "Missing arguments after 'pip install'", self._node
                )
            ]
        problems = parent.register_library(*argv[2:])  # Skipping %pip install
        return [problem.from_node(problem.code, problem.message, self._node) for problem in problems]

    # Cache re-used regex (and ensure issues are raised during class init instead of upon first use).
    _splitter = re.compile(r"(?<!\\)\n")

    @classmethod
    def _split(cls, code: str) -> list[str]:
        """Split pip cell code into multiple arguments

        Note:
            PipCell should be a pip command, i.e. single line possible spanning multilines escaped with backslashes.

        Sources:
            https://docs.databricks.com/en/libraries/notebooks-python-libraries.html#manage-libraries-with-pip-commands
        """
        # strip preliminary comments
        pip_idx = code.find("pip")
        if pip_idx > 0 and code[pip_idx - 1] in {'%', '!'}:
            pip_idx -= 1
        code = code[pip_idx:]
        # look for standalone '\n'
        match = cls._splitter.search(code)
        if match:
            code = code[: match.start()]  # Remove code after non-escaped newline
        # make single line
        code = code.replace("\\\n", " ")
        return shlex.split(code, posix=True)



import functools
import json
import logging
import re
from abc import abstractmethod
from collections.abc import Iterable, Collection
from dataclasses import dataclass
from datetime import timedelta

from databricks.labs.blueprint.limiter import rate_limited
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors.platform import (
    BadRequest,
    DeadlineExceeded,
    InternalError,
    NotFound,
    ResourceConflict,
)
from databricks.sdk.retries import retried
from databricks.sdk.service import iam
from databricks.sdk.service.iam import Group, User

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)

SYSTEM_GROUPS: list[str] = ["users", "admins", "account users"]


@dataclass
class MigratedGroup:
    id_in_workspace: str
    name_in_workspace: str
    name_in_account: str
    temporary_name: str
    members: str | None = None
    entitlements: str | None = None
    external_id: str | None = None
    roles: str | None = None

    @classmethod
    def partial_info(cls, workspace: iam.Group, account: iam.Group):
        """This method is only intended for use in tests"""
        assert workspace.id is not None
        assert workspace.display_name is not None
        assert account.display_name is not None
        return cls(
            id_in_workspace=workspace.id,
            name_in_workspace=workspace.display_name,
            name_in_account=account.display_name,
            temporary_name=f"tmp-{workspace.display_name}",
            external_id=workspace.external_id,
        )

    def decode_members(self):
        return [iam.ComplexValue.from_dict(_) for _ in json.loads(self.members)]


class MigrationState:
    """Holds migration state of workspace-to-account groups"""

    def __init__(self, groups: list[MigratedGroup]):
        self._name_to_group: dict[str, MigratedGroup] = {_.name_in_workspace: _ for _ in groups}
        self._id_to_group: dict[str, MigratedGroup] = {_.id_in_workspace: _ for _ in groups}
        self.groups: list[MigratedGroup] = groups

    def get_target_principal(self, name: str) -> str | None:
        migrated_group = self._name_to_group.get(name)
        if migrated_group is None:
            return None
        return migrated_group.name_in_account

    def get_temp_principal(self, name: str) -> str | None:
        migrated_group = self._name_to_group.get(name)
        if migrated_group is None:
            return None
        return migrated_group.temporary_name

    def is_in_scope(self, name: str) -> bool:
        if name is None:
            return False
        return name in self._name_to_group

    def __len__(self):
        return len(self._name_to_group)

    def apply_to_renamed_groups(self, ws: WorkspaceClient) -> bool:
        """(Production) Apply migration state to groups that have been renamed in the workspace."""
        return self._apply_to_groups(ws, renamed=True)

    def apply_to_groups_with_different_names(self, ws: WorkspaceClient) -> bool:
        """(Integration Testing) Apply to groups that have different names in the workspace and account."""
        return self._apply_to_groups(ws, renamed=False)

    def _apply_to_groups(self, ws: WorkspaceClient, *, renamed: bool = False) -> bool:
        if len(self) == 0:
            logger.info("No valid groups selected, nothing to do.")
            return True
        logger.info(f"Migrating permissions for {len(self)} account groups.")
        total_permissions = 0
        success_groups = 0
        errors: list[Exception] = []
        for migrated_group in self.groups:
            name_in_workspace = migrated_group.name_in_workspace
            if renamed:
                # during integration testing, local and account group have completely different names,
                # that simplifies visual debuggability. In production, we need to first rename local
                # groups, otherwise we get `Workspace group XXX is not a workspace local group` and
                # the migration fails.
                name_in_workspace = migrated_group.temporary_name
            name_in_account = migrated_group.name_in_account
            try:
                group_permissions = self._migrate_group_permissions_paginated(ws, name_in_workspace, name_in_account)
                logger.info(
                    f"Migrated {group_permissions} permissions: {name_in_workspace} (workspace) -> {name_in_account} (account)"
                )
                total_permissions += group_permissions
                success_groups += 1
            except IOError as e:
                logger.error(f"failed-group-migration: {name_in_workspace} -> {name_in_account}: {e}")
                errors.append(e)
        logger.info(f"Migrated {total_permissions} permissions for {success_groups}/{len(self)} groups successfully.")
        if errors:
            logger.error(f"Migrating permissions failed for {len(errors)}/{len(self)} groups.")
            raise ManyError(errors)
        return True

    @staticmethod
    def _migrate_group_permissions_paginated(ws: WorkspaceClient, name_in_workspace: str, name_in_account: str) -> int:
        batch_size = 1000
        logger.info(f"Migrating permissions: {name_in_workspace} (workspace) -> {name_in_account} (account) starting")
        permissions_migrated = 0
        while True:
            result = ws.permission_migration.migrate_permissions(
                ws.get_workspace_id(),
                name_in_workspace,
                name_in_account,
                size=batch_size,
            )
            if not result.permissions_migrated:
                logger.info(
                    f"Migrating permissions: {name_in_workspace} (workspace) -> {name_in_account} (account) finished"
                )
                return permissions_migrated
            permissions_migrated += result.permissions_migrated
            logger.info(
                f"Migrating permissions: {name_in_workspace} (workspace) -> {name_in_account} (account) progress={permissions_migrated}(+{result.permissions_migrated})"
            )


class GroupMigrationStrategy:
    def __init__(
        self,
        workspace_groups_in_workspace,
        account_groups_in_account,
        /,
        renamed_groups_prefix,
        include_group_names=None,
    ):
        self.renamed_groups_prefix = renamed_groups_prefix
        self.workspace_groups_in_workspace = workspace_groups_in_workspace
        self.account_groups_in_account = account_groups_in_account
        self.include_group_names = include_group_names

    @abstractmethod
    def generate_migrated_groups(self) -> Iterable[MigratedGroup]:
        raise NotImplementedError

    def get_filtered_groups(self):
        if not self.include_group_names:
            logger.info("No group listing provided, all matching groups will be migrated")
            return self.workspace_groups_in_workspace
        logger.info("Group listing provided, a subset of all groups will be migrated")
        return {
            group_name: self.workspace_groups_in_workspace[group_name]
            for group_name in self.workspace_groups_in_workspace.keys()
            if group_name in self.include_group_names
        }

    @staticmethod
    def _safe_match(group_name: str, pattern: re.Pattern) -> str:
        try:
            match = pattern.search(group_name)
            if not match:
                return group_name
            match_groups = match.groups()
            if match_groups:
                return match_groups[0]
            return match.group()
        except re.error:
            return group_name

    @staticmethod
    def _safe_sub(group_name: str, pattern: re.Pattern, replace: str) -> str:
        try:
            return pattern.sub(replace, group_name)
        except re.error:
            logger.warning(f"Failed to apply Regex Expression {pattern} on Group Name {group_name}")
            return group_name


class MatchingNamesStrategy(GroupMigrationStrategy):
    def __init__(
        self,
        workspace_groups_in_workspace,
        account_groups_in_account,
        *,
        renamed_groups_prefix: str,
        include_group_names: list[str] | None,
    ):
        super().__init__(
            workspace_groups_in_workspace,
            account_groups_in_account,
            include_group_names=include_group_names,
            renamed_groups_prefix=renamed_groups_prefix,
        )

    def generate_migrated_groups(self) -> Iterable[MigratedGroup]:
        workspace_groups = self.get_filtered_groups()
        for group in workspace_groups.values():
            account_group = self.account_groups_in_account.get(group.display_name)
            if not account_group:
                logger.info(
                    f"Couldn't find a matching account group for {group.display_name} group using name matching"
                )
                continue
            temporary_name = f"{self.renamed_groups_prefix}{group.display_name}"
            yield MigratedGroup(
                id_in_workspace=group.id,
                name_in_workspace=group.display_name,
                name_in_account=group.display_name,
                temporary_name=temporary_name,
                external_id=account_group.external_id,
                members=json.dumps([gg.as_dict() for gg in group.members]) if group.members else None,
                roles=json.dumps([gg.as_dict() for gg in group.roles]) if group.roles else None,
                entitlements=json.dumps([gg.as_dict() for gg in group.entitlements]) if group.entitlements else None,
            )


class MatchByExternalIdStrategy(GroupMigrationStrategy):
    def __init__(
        self,
        workspace_groups_in_workspace,
        account_groups_in_account,
        *,
        renamed_groups_prefix: str,
        include_group_names: list[str] | None,
    ):
        super().__init__(
            workspace_groups_in_workspace,
            account_groups_in_account,
            include_group_names=include_group_names,
            renamed_groups_prefix=renamed_groups_prefix,
        )

    def generate_migrated_groups(self) -> Iterable[MigratedGroup]:
        workspace_groups = self.get_filtered_groups()
        account_groups_by_id = {group.external_id: group for group in self.account_groups_in_account.values()}
        for group in workspace_groups.values():
            account_group = account_groups_by_id.get(group.external_id)
            if not account_group:
                logger.info(f"Couldn't find a matching account group for {group.display_name} group with external_id")
                continue
            temporary_name = f"{self.renamed_groups_prefix}{group.display_name}"
            yield MigratedGroup(
                id_in_workspace=group.id,
                name_in_workspace=group.display_name,
                name_in_account=account_group.display_name,
                temporary_name=temporary_name,
                external_id=account_group.external_id,
                members=json.dumps([gg.as_dict() for gg in group.members]) if group.members else None,
                roles=json.dumps([gg.as_dict() for gg in group.roles]) if group.roles else None,
                entitlements=(json.dumps([gg.as_dict() for gg in group.entitlements]) if group.entitlements else None),
            )


class RegexSubStrategy(GroupMigrationStrategy):
    def __init__(
        self,
        workspace_groups_in_workspace: dict[str, Group],
        account_groups_in_account: dict[str, Group],
        *,
        renamed_groups_prefix: str,
        include_group_names: list[str] | None,
        workspace_group_regex: str,
        workspace_group_replace: str,
    ):
        super().__init__(
            workspace_groups_in_workspace,
            account_groups_in_account,
            include_group_names=include_group_names,
            renamed_groups_prefix=renamed_groups_prefix,
        )
        self.workspace_group_regex = workspace_group_regex  # Keep to support legacy public API
        self.workspace_group_replace = workspace_group_replace

        self._workspace_group_pattern = re.compile(self.workspace_group_regex)

    def generate_migrated_groups(self) -> Iterable[MigratedGroup]:
        workspace_groups = self.get_filtered_groups()
        for group in workspace_groups.values():
            name_in_account = self._safe_sub(
                group.display_name,
                self._workspace_group_pattern,
                self.workspace_group_replace,
            )
            account_group = self.account_groups_in_account.get(name_in_account)
            if not account_group:
                logger.info(
                    f"Couldn't find a matching account group for {group.display_name} group with regex substitution"
                )
                continue
            temporary_name = f"{self.renamed_groups_prefix}{group.display_name}"
            yield MigratedGroup(
                id_in_workspace=group.id,
                name_in_workspace=group.display_name,
                name_in_account=name_in_account,
                temporary_name=temporary_name,
                external_id=account_group.external_id,
                members=json.dumps([gg.as_dict() for gg in group.members]) if group.members else None,
                roles=json.dumps([gg.as_dict() for gg in group.roles]) if group.roles else None,
                entitlements=json.dumps([gg.as_dict() for gg in group.entitlements]) if group.entitlements else None,
            )


class RegexMatchStrategy(GroupMigrationStrategy):
    def __init__(
        self,
        workspace_groups_in_workspace,
        account_groups_in_account,
        *,
        renamed_groups_prefix: str,
        include_group_names: list[str] | None,
        workspace_group_regex: str,
        account_group_regex: str,
    ):
        super().__init__(
            workspace_groups_in_workspace,
            account_groups_in_account,
            include_group_names=include_group_names,
            renamed_groups_prefix=renamed_groups_prefix,
        )
        # Keep to support legacy public API
        self.workspace_group_regex = workspace_group_regex
        self.account_group_regex = account_group_regex

        self._workspace_group_pattern = re.compile(self.workspace_group_regex)
        self._account_group_pattern = re.compile(self.account_group_regex)

    def generate_migrated_groups(self) -> Iterable[MigratedGroup]:
        workspace_groups_by_match = {
            self._safe_match(group_name, self._workspace_group_pattern): group
            for group_name, group in self.get_filtered_groups().items()
        }
        account_groups_by_match = {
            self._safe_match(group_name, self._account_group_pattern): group
            for group_name, group in self.account_groups_in_account.items()
        }
        for group_match, ws_group in workspace_groups_by_match.items():
            account_group = account_groups_by_match.get(group_match)
            if not account_group:
                logger.info(
                    f"Couldn't find a matching account group for {ws_group.display_name} group with regex matching"
                )
                continue
            temporary_name = f"{self.renamed_groups_prefix}{ws_group.display_name}"
            yield MigratedGroup(
                id_in_workspace=ws_group.id,
                name_in_workspace=ws_group.display_name,
                name_in_account=account_group.display_name,
                temporary_name=temporary_name,
                external_id=account_group.external_id,
                members=json.dumps([gg.as_dict() for gg in ws_group.members]) if ws_group.members else None,
                roles=json.dumps([gg.as_dict() for gg in ws_group.roles]) if ws_group.roles else None,
                entitlements=(
                    json.dumps([gg.as_dict() for gg in ws_group.entitlements]) if ws_group.entitlements else None
                ),
            )


class GroupDeletionIncompleteError(RuntimeError):
    __slots__ = ("group_id", "display_name")

    def __init__(self, group_id: str, display_name: str | None) -> None:
        msg = f"Group deletion incomplete: {display_name if display_name else '<name-missing>'} (id={group_id})"
        super().__init__(msg)
        self.group_id = group_id
        self.display_name = display_name


class GroupRenameIncompleteError(RuntimeError):
    __slots__ = ("group_id", "old_name", "new_name")

    def __init__(self, group_id: str, old_name: str, new_name: str) -> None:
        super().__init__(f"Rename incomplete for group {group_id}: {old_name} -> {new_name}")
        self.group_id = group_id
        self.old_name = old_name
        self.new_name = new_name


class GroupManager(CrawlerBase[MigratedGroup]):

    def __init__(  # pylint: disable=too-many-arguments
        self,
        sql_backend: SqlBackend,
        ws: WorkspaceClient,
        inventory_database: str,
        include_group_names: list[str] | None = None,
        renamed_group_prefix: str | None = "db-temp-",
        workspace_group_regex: str | None = None,
        workspace_group_replace: str | None = None,
        account_group_regex: str | None = None,
        verify_timeout: timedelta | None = timedelta(minutes=2),
        *,
        external_id_match: bool = False,
    ):
        super().__init__(sql_backend, "hive_metastore", inventory_database, "groups", MigratedGroup)
        if not renamed_group_prefix:
            renamed_group_prefix = "db-temp-"

        self._ws = ws
        self._include_group_names = include_group_names
        self._renamed_group_prefix = renamed_group_prefix
        self._workspace_group_regex = workspace_group_regex
        self._workspace_group_replace = workspace_group_replace
        self._account_group_regex = account_group_regex
        self._external_id_match = external_id_match
        self._verify_timeout = verify_timeout
        self._account_groups_lookup = AccountGroupLookup(ws)

    def rename_groups(self):
        account_groups_in_workspace = self._account_groups_in_workspace()
        workspace_groups_in_workspace = self._workspace_groups_in_workspace()
        # Renaming a group is eventually consistent, and not monotonically consistent, with a rather long time to
        # converge: internally the Databricks API caches some things for up to 60s. To avoid excessive wait times when
        # large numbers of groups need to be deleted (some deployments have >10K groups) we use the following steps:
        #  1. Rename all the groups.
        #  2. Confirm for each group that direct GETs yield the new name.
        #  3. Confirm that group enumeration no longer includes the deleted groups.
        # This caution is necessary because otherwise downstream tasks (like reflect_account_groups_on_workspace()) may
        # skip a renamed group because it doesn't appear to be present.
        groups_to_migrate = self.get_migration_state().groups
        rename_tasks = []
        waiting_tasks = []
        renamed_groups = []
        logger.info(f"Starting to rename {len(groups_to_migrate)} groups for migration...")
        for migrated_group in groups_to_migrate:
            if migrated_group.name_in_account in account_groups_in_workspace:
                logger.info(f"Skipping {migrated_group.name_in_account}: already in workspace")
                continue
            if migrated_group.temporary_name in workspace_groups_in_workspace:
                logger.info(f"Skipping {migrated_group.name_in_workspace}: already renamed")
                continue
            rename_tasks.append(
                functools.partial(
                    self._rename_group,
                    migrated_group.id_in_workspace,
                    migrated_group.name_in_workspace,
                    migrated_group.temporary_name,
                )
            )
            waiting_tasks.append(
                functools.partial(
                    self._wait_for_group_rename,
                    migrated_group.id_in_workspace,
                    migrated_group.name_in_workspace,
                    migrated_group.temporary_name,
                )
            )
            renamed_groups.append((migrated_group.id_in_workspace, migrated_group.temporary_name))
        # Step 1: Rename all the groups.
        _, errors = Threads.gather("rename groups in the workspace", rename_tasks)
        if errors:
            logger.error(f"During renaming of workspace groups {len(errors)} errors occurred. See debug logs.")
            raise ManyError(errors)
        # Step 2: Confirm that direct GETs yield the updated information.
        _, errors = Threads.gather("waiting for renamed groups in the workspace", waiting_tasks)
        if errors:
            logger.error(f"While waiting for renamed workspace groups {len(errors)} errors occurred. See debug logs.")
            raise ManyError(errors)
        # Step 3: Wait for enumeration to also reflect the updated information.
        self._wait_for_renamed_groups(renamed_groups)

    def current_user_in_owner_group(self, group_name: str) -> bool:
        return self._account_groups_lookup.user_in_group(group_name, self._ws.current_user.me())

    def _rename_group(self, group_id: str, old_group_name: str, new_group_name: str) -> None:
        logger.debug(f"Renaming group: {old_group_name} (id={group_id}) -> {new_group_name}")
        self._rate_limited_rename_group_with_retry(group_id, new_group_name)

    @retried(on=[InternalError, ResourceConflict, DeadlineExceeded])
    @rate_limited(max_requests=10, burst_period_seconds=60)
    def _rate_limited_rename_group_with_retry(self, group_id: str, new_group_name: str) -> None:
        ops = [iam.Patch(iam.PatchOp.REPLACE, "displayName", new_group_name)]
        self._ws.groups.patch(group_id, operations=ops)

    @retried(on=[GroupRenameIncompleteError], timeout=timedelta(seconds=90))
    def _wait_for_group_rename(self, group_id: str, old_group_name: str, new_group_name: str) -> None:
        # The groups API is eventually consistent, but not monotonically consistent. Here we verify that the effects of
        # the rename have taken effect, and try to compensate for the lack of monotonic consistency by requiring two
        # subsequent calls to confirm the rename. REST API internals cache things for up to 60s, and we see times close
        # to this during testing. The retry timeout reflects this: if it's taking much longer then something else is
        # wrong.
        self._check_group_rename(group_id, old_group_name, new_group_name, logging.DEBUG)
        self._check_group_rename(group_id, old_group_name, new_group_name, logging.WARNING)
        logger.debug(f"Group rename is assumed complete: {old_group_name} (id={group_id}) -> {new_group_name}")

    def _check_group_rename(self, group_id, old_group_name: str, new_group_name: str, pending_log_level: int) -> None:
        group = self._ws.groups.get(group_id)
        if group.display_name == old_group_name:
            logger.log(
                pending_log_level,
                f"Group still has old name; still waiting for rename to take effect: {old_group_name} (id={group_id}) -> {new_group_name}",
            )
            raise GroupRenameIncompleteError(group_id, old_group_name, new_group_name)
        if group.display_name != new_group_name:
            # Group has an entirely unexpected name; something else is interfering.
            msg = f"While waiting for group rename ({old_group_name} (id={group_id}) -> {new_group_name}) an unexpected name was observed: {group.display_name}"
            raise RuntimeError(msg)
        logger.debug(f"Group rename has possibly taken effect: {old_group_name} (id={group_id}) -> {new_group_name}")

    @retried(on=[ManyError], timeout=timedelta(minutes=2))
    def _wait_for_renamed_groups(self, expected_groups: Collection[tuple[str, str]]) -> None:
        # The groups API is eventually consistent, but not monotonically consistent. Here we verify that the group
        # has been deleted, and try to compensate for the lack of monotonic consistency by requiring two subsequent
        # calls to confirm deletion. REST API internals cache things for up to 60s, and we see times close to this
        # during testing. The retry timeout reflects this: if it's taking much longer then something else is wrong.
        self._check_for_renamed_groups(expected_groups, logging.DEBUG)
        self._check_for_renamed_groups(expected_groups, logging.WARNING)
        logger.debug(f"Group enumeration showed all {len(expected_groups)} renamed groups; assuming complete.")

    def _check_for_renamed_groups(self, expected_groups: Collection[tuple[str, str]], pending_log_level: int) -> None:
        attributes = "id,displayName"
        found_groups = {
            group.id: group.display_name
            for group in self._list_workspace_groups("WorkspaceGroup", attributes)
            if group.display_name
        }
        pending_renames: list[RuntimeError] = []
        for group_id, expected_name in expected_groups:
            found_name = found_groups.get(group_id, None)
            if found_name is None:
                logger.warning(f"Group enumeration omits renamed group: {group_id} (renamed to {expected_name})")
                pending_renames.append(RuntimeError(f"Missing group with id: {group_id} (renamed to {expected_name})"))
            elif found_name != expected_name:
                logger.log(
                    pending_log_level,
                    f"Group enumeration does not yet reflect rename: {group_id} (renamed to {expected_name} but currently {found_name})",
                )
                pending_renames.append(GroupRenameIncompleteError(group_id, found_name, expected_name))
            else:
                logger.debug(f"Group enumeration reflects renamed group: {group_id} (renamed to {expected_name})")
        if pending_renames:
            raise ManyError(pending_renames)

    def reflect_account_groups_on_workspace(self):
        tasks = []
        account_groups_in_account = self._account_groups_lookup.get_mapping()
        account_groups_in_workspace = self._account_groups_in_workspace()
        workspace_groups_in_workspace = self._workspace_groups_in_workspace()
        groups_to_migrate = self.get_migration_state().groups
        logger.info(f"Starting to reflect {len(groups_to_migrate)} account groups into workspace for migration...")
        for migrated_group in groups_to_migrate:
            if migrated_group.name_in_account in account_groups_in_workspace:
                logger.info(f"Skipping {migrated_group.name_in_account}: already in workspace")
                continue
            if migrated_group.name_in_account in workspace_groups_in_workspace:
                logger.error(f"Skipping {migrated_group.name_in_account}: group already exists in workspace")
                continue
            if migrated_group.name_in_account not in account_groups_in_account:
                logger.warning(f"Skipping {migrated_group.name_in_account}: not in account")
                continue
            group_id = account_groups_in_account[migrated_group.name_in_account].id
            tasks.append(functools.partial(self._reflect_account_group_to_workspace, group_id))
        _, errors = Threads.gather("reflect account groups on this workspace", tasks)
        if len(errors) > 0:
            raise ManyError(errors)

    def get_migration_state(self) -> MigrationState:
        return MigrationState(list(self.snapshot()))

    def delete_original_workspace_groups(self):
        account_groups_in_workspace = self._account_groups_in_workspace()
        migrated_groups = self.snapshot()
        logger.info(f"Starting to remove {len(migrated_groups)} migrated workspace groups...")
        # Group deletion is eventually consistent, and not monotonically consistent, with a rather long time to
        # converge: internally the Databricks API caches some things for up to 60s. To avoid excessive wait times when
        # large numbers of groups need to be deleted (some deployments have >10K groups) we use the following steps:
        #  1. Delete all the groups.
        #  2. Confirm for each group that direct GETs no longer see the group.
        #  3. Confirm that group enumeration no longer includes the deleted groups.
        deletion_tasks = []
        waiting_tasks = []
        deleted_groups = []
        for migrated_group in migrated_groups:
            if migrated_group.name_in_account not in account_groups_in_workspace:
                logger.warning(
                    f"Not deleting group {migrated_group.temporary_name}(id={migrated_group.id_in_workspace}) (originally {migrated_group.name_in_workspace}): its migrated account group ({migrated_group.name_in_account}) cannot be found."
                )
                continue
            deletion_tasks.append(
                functools.partial(
                    self._delete_workspace_group, migrated_group.id_in_workspace, migrated_group.temporary_name
                )
            )
            waiting_tasks.append(
                functools.partial(
                    self._wait_for_workspace_group_deletion,
                    migrated_group.id_in_workspace,
                    migrated_group.temporary_name,
                )
            )
            deleted_groups.append(migrated_group)
        # Step 1: Delete the groups.
        _, errors = Threads.gather("removing original workspace groups", deletion_tasks)
        if errors:
            logger.error(f"During deletion of workspace groups got {len(errors)} errors. See debug logs.")
            raise ManyError(errors)
        # Step 2: Confirm that direct GETs no longer return the deleted group.
        _, errors = Threads.gather("waiting for removal of original workspace groups", waiting_tasks)
        if errors:
            logger.error(f"Waiting for deletion of workspace groups got {len(errors)} errors. See debug logs.")
            raise ManyError(errors)
        # Step 3: Confirm that enumeration no longer returns the deleted groups.
        self._wait_for_deleted_workspace_groups(deleted_groups)

    def _try_fetch(self) -> Iterable[MigratedGroup]:
        state = []
        for row in self._sql_backend.fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            state.append(MigratedGroup(*row))

        if not self._include_group_names:
            return state

        new_state = []
        group_name_with_state = {migrated_group.name_in_workspace: migrated_group for migrated_group in state}
        for group_name in self._include_group_names:
            if group_name in group_name_with_state:
                new_state.append(group_name_with_state[group_name])
            else:
                logger.warning(
                    f"Group {group_name} defined in configuration does not exist on the groups table. "
                    "Consider checking if the group exist in the workspace or re-running the assessment."
                )
        return new_state

    def _crawl(self) -> Iterable[MigratedGroup]:
        workspace_groups_in_workspace = self._workspace_groups_in_workspace()
        account_groups_in_account = self._account_groups_lookup.get_mapping()
        strategy = self._get_strategy(workspace_groups_in_workspace, account_groups_in_account)
        yield from strategy.generate_migrated_groups()

    def validate_group_membership(self) -> list[dict]:
        workspace_groups_in_workspace = self._workspace_groups_in_workspace()
        account_groups_in_account = self._account_groups_lookup.get_mapping()
        strategy = self._get_strategy(workspace_groups_in_workspace, account_groups_in_account)
        migrated_groups = list(strategy.generate_migrated_groups())
        mismatch_group = []
        retry_on_internal_error = retried(on=[InternalError], timeout=self._verify_timeout)
        get_account_group = retry_on_internal_error(self._get_account_group)
        logger.info(f"Starting to validate {len(migrated_groups)} migrated workspace groups...")
        for ws_group in migrated_groups:
            # Users with the same display name but different email will be deduplicated!
            ws_members_set = {m.get("display") for m in json.loads(ws_group.members)} if ws_group.members else set()
            acc_group_id = account_groups_in_account[ws_group.name_in_account].id
            acc_group = get_account_group(acc_group_id)
            if not acc_group:
                logger.debug(
                    f"Skipping validation; account group no longer present: {ws_group.name_in_account} (id={acc_group_id})"
                )
                continue  # group not present anymore
            acc_members_set = {a.as_dict().get("display") for a in acc_group.members} if acc_group.members else set()
            set_diff = (ws_members_set - acc_members_set).union(acc_members_set - ws_members_set)
            if not set_diff:
                logger.debug(f"Validated group, no differences found: {ws_group.name_in_account} (id={acc_group_id})")
                continue
            mismatch_group.append(
                {
                    "wf_group_name": ws_group.name_in_workspace,
                    "wf_group_members_count": len(ws_members_set),
                    "acc_group_name": ws_group.name_in_account,
                    "acc_group_members_count": len(acc_members_set),
                    "group_members_difference": len(ws_members_set) - len(acc_members_set),
                }
            )
        if not mismatch_group:
            logger.info("There are no groups with different membership between account and workspace.")
        else:
            logger.info(
                f"There are {len(mismatch_group)} (of {len(migrated_groups)}) groups with different membership between account and workspace."
            )
        return mismatch_group

    def has_workspace_group(self, name):
        groups = self._workspace_groups_in_workspace()
        return name in groups

    def has_account_group(self, name):
        groups = self._account_groups_in_workspace()
        return name in groups

    def _workspace_groups_in_workspace(self) -> dict[str, Group]:
        attributes = "id,displayName,meta,externalId,members,roles,entitlements"
        groups = {}
        for group in self._list_workspace_groups("WorkspaceGroup", attributes):
            if not group.display_name:
                logger.debug(f"Ignoring workspace group without name: {group.id}")
                continue
            groups[group.display_name] = group
        return groups

    def _account_groups_in_workspace(self) -> dict[str, Group]:
        groups = {}
        for group in self._list_workspace_groups("Group", "id,displayName,externalId,meta"):
            if not group.display_name:
                logger.debug(f"Ignoring account group in workspace without name: {group.id}")
                continue
            groups[group.display_name] = group
        return groups

    @staticmethod
    def _is_group_out_of_scope(group: iam.Group, resource_type: str) -> bool:
        if group.display_name in SYSTEM_GROUPS:
            return True
        meta = group.meta
        if not meta:
            return False
        if meta.resource_type != resource_type:
            return True
        return False

    def _list_workspace_groups(self, resource_type: str, scim_attributes: str) -> list[iam.Group]:
        results = []
        logger.info(f"Listing workspace groups (resource_type={resource_type}) with {scim_attributes} ...")
        # If members are requested during enumeration the API can time out. In this case we fall back on
        # a strategy of enumerating the bare minimum and request full attributes for each group individually.
        attributes = scim_attributes.split(",")
        if "members" in attributes:
            retry_on_internal_error = retried(on=[InternalError], timeout=self._verify_timeout)
            get_group = retry_on_internal_error(self._get_group)
            # Limit to the attributes we need for determining if the group is out of scope; the rest are fetched later.
            scan_attributes = [attribute for attribute in attributes if attribute in {"id", "displayName", "meta"}]
            for group in self._ws.groups.list(attributes=",".join(scan_attributes)):
                if self._is_group_out_of_scope(group, resource_type):
                    continue
                group_with_all_attributes = get_group(group.id)
                if not group_with_all_attributes:
                    continue
                results.append(group_with_all_attributes)
        else:
            for group in self._ws.groups.list(attributes=scim_attributes):
                if self._is_group_out_of_scope(group, resource_type):
                    continue
                results.append(group)
        logger.info(f"Found {len(results)} {resource_type}")
        return results

    @rate_limited(max_requests=255, burst_period_seconds=60)
    def _get_group(self, group_id: str) -> iam.Group | None:
        try:
            return self._ws.groups.get(group_id)
        except NotFound:
            # during integration tests, we may get certain groups removed,
            # which will cause timeout errors because of groups no longer there.
            return None

    @rate_limited(max_requests=20)
    def _get_account_group(self, group_id: str) -> Group | None:
        try:
            raw = self._ws.api_client.do("GET", f"/api/2.0/account/scim/v2/Groups/{group_id}")
            return iam.Group.from_dict(raw)  # type: ignore[arg-type]
        except NotFound:
            # the given group has been removed from the account after getting the group and before running this method
            logger.warning(f"Group with ID {group_id} does not exist anymore in the Databricks account.")
            return None

    def _delete_workspace_group_and_wait_for_deletion(self, group_id: str, display_name: str) -> str:
        logger.debug(f"Deleting workspace group: {display_name} (id={group_id})")
        self._delete_workspace_group(group_id, display_name)
        logger.debug(f"Waiting for workspace group deletion to take effect: {display_name} (id={group_id})")
        self._wait_for_workspace_group_deletion(group_id, display_name)
        return group_id

    @retried(on=[InternalError, ResourceConflict, DeadlineExceeded])
    @rate_limited(max_requests=35, burst_period_seconds=60)
    def _rate_limited_group_delete_with_retry(self, group_id: str) -> None:
        try:
            self._ws.groups.delete(id=group_id)
        except NotFound:
            pass

    def _delete_workspace_group(self, group_id: str, display_name: str) -> None:
        logger.debug(f"Deleting workspace group: {display_name} (id={group_id})")
        self._rate_limited_group_delete_with_retry(group_id)

    @retried(on=[GroupDeletionIncompleteError], timeout=timedelta(seconds=90))
    def _wait_for_workspace_group_deletion(self, group_id: str, display_name: str) -> None:
        # The groups API is eventually consistent, but not monotonically consistent. Here we verify that the group
        # has been deleted, and try to compensate for the lack of monotonic consistency by requiring two subsequent
        # calls to confirm deletion. REST API internals cache things for up to 60s, and we see times close to this
        # during testing. The retry timeout reflects this: if it's taking much longer then something else is wrong.
        self._check_workspace_group_deletion(group_id, display_name, logging.DEBUG)
        self._check_workspace_group_deletion(group_id, display_name, logging.WARNING)
        logger.debug(f"Workspace group is assumed deleted: {display_name} (id={group_id})")

    def _check_workspace_group_deletion(self, group_id: str, display_name: str, pending_log_level: int) -> None:
        try:
            _ = self._ws.groups.get(id=group_id)
            logger.log(
                pending_log_level,
                f"Deleted group is still present; still waiting for deletion to take effect: {display_name} (id={group_id})",
            )
            # Deletion is still pending.
            raise GroupDeletionIncompleteError(group_id, display_name)
        except NotFound:
            logger.debug(f"Workspace group not found; possibly deleted: {display_name} (id={group_id})")

    @retried(on=[ManyError], timeout=timedelta(minutes=5))
    def _wait_for_deleted_workspace_groups(self, deleted_workspace_groups: list[MigratedGroup]) -> None:
        # The groups API is eventually consistent, but not monotonically consistent. Here we verify that enumerating
        # all groups no longer includes the deleted groups. We try to compensate for the lack of monotonic consistency
        # by requiring two subsequent enumerations to omit all deleted groups. REST API internals cache things for up
        # to 60s. The retry timeout reflects this, and the fact that enumeration can take a long time for large numbers
        # of groups. (Currently there is no way to configure the retry handler to retry at least once, so the timeout
        # needs to be high enough to allow at least one retry.)
        self._check_for_deleted_workspace_groups(deleted_workspace_groups, logging.DEBUG)
        self._check_for_deleted_workspace_groups(deleted_workspace_groups, logging.WARNING)
        logger.debug(
            f"Group enumeration omitted all {len(deleted_workspace_groups)} workspace groups; assuming deleted."
        )

    def _check_for_deleted_workspace_groups(
        self, deleted_workspace_groups: list[MigratedGroup], still_present_log_level: int
    ) -> None:
        attributes = "id,displayName"
        expected_deletions = {group.id_in_workspace for group in deleted_workspace_groups}
        pending_deletions = []
        for group in self._list_workspace_groups("WorkspaceGroup", attributes):
            if group.id in expected_deletions:
                pending_deletions.append(GroupDeletionIncompleteError(group.id, group.display_name))
        if pending_deletions:
            if logger.isEnabledFor(still_present_log_level):
                logger.log(
                    still_present_log_level,
                    f"Group enumeration still contains {len(pending_deletions)}/{len(expected_deletions)} deleted workspace groups.",
                )
                for pending_deletion in pending_deletions:
                    logger.log(
                        still_present_log_level,
                        f"Group enumeration still contains deleted group: {pending_deletion.display_name}(id={pending_deletion.group_id})",
                    )
            raise ManyError(pending_deletions)
        logger.debug(
            f"Group enumeration does not contain any of the {len(expected_deletions)} deleted workspace groups; possibly deleted."
        )

    @retried(on=[InternalError, ResourceConflict, DeadlineExceeded])
    @rate_limited(max_requests=5)
    def _reflect_account_group_to_workspace(self, account_group_id: str):
        try:
            # TODO: add OpenAPI spec for it
            path = f"/api/2.0/preview/permissionassignments/principals/{account_group_id}"
            self._ws.api_client.do("PUT", path, data=json.dumps({"permissions": ["USER"]}))
            return True
        except BadRequest:
            # already exists
            return True
        except NotFound:
            # the given group has been removed from the account after getting the group and before running this method
            logger.warning(f"Group with ID {account_group_id} does not exist anymore in the Databricks account.")
            return True

    def _get_strategy(
        self, workspace_groups_in_workspace: dict[str, Group], account_groups_in_account: dict[str, Group]
    ) -> GroupMigrationStrategy:
        if self._workspace_group_regex is not None and self._workspace_group_replace is not None:
            return RegexSubStrategy(
                workspace_groups_in_workspace,
                account_groups_in_account,
                renamed_groups_prefix=self._renamed_group_prefix,
                include_group_names=self._include_group_names,
                workspace_group_regex=self._workspace_group_regex,
                workspace_group_replace=self._workspace_group_replace,
            )
        if self._workspace_group_regex is not None and self._account_group_regex is not None:
            return RegexMatchStrategy(
                workspace_groups_in_workspace,
                account_groups_in_account,
                renamed_groups_prefix=self._renamed_group_prefix,
                include_group_names=self._include_group_names,
                workspace_group_regex=self._workspace_group_regex,
                account_group_regex=self._account_group_regex,
            )
        if self._external_id_match:
            return MatchByExternalIdStrategy(
                workspace_groups_in_workspace,
                account_groups_in_account,
                renamed_groups_prefix=self._renamed_group_prefix,
                include_group_names=self._include_group_names,
            )
        return MatchingNamesStrategy(
            workspace_groups_in_workspace,
            account_groups_in_account,
            renamed_groups_prefix=self._renamed_group_prefix,
            include_group_names=self._include_group_names,
        )


class AccountGroupLookup:
    def __init__(self, ws: WorkspaceClient):
        self._ws = ws

    def pick_owner_group(self, prompt: Prompts) -> str | None:
        # This method is used to select the group that will be used as the owner group.
        # The owner group will be assigned by default to all migrated tables/schemas
        user_id = self._ws.current_user.me().id
        if not user_id:
            logger.error("Couldn't find the user id of the current user.")
            return None
        groups = self._user_account_groups(user_id)
        if not groups:
            logger.warning("No account groups found for the current user.")
            return None
        if len(groups) == 1:
            return groups[0].display_name
        group_names = [group.display_name for group in groups]
        return prompt.choice("Select the group to be used as the owner group", group_names, max_attempts=3)

    def user_in_group(self, group_name: str, user: User) -> bool:
        # This method is used to validate that the current user is a member of the group
        user_id = user.id
        if not user_id:
            logger.warning("No user found for the current session.")
            return False
        groups = self._user_account_groups(user_id)
        if not groups:
            logger.warning("No account groups found for the current user.")
            return False
        group_names = [group.display_name for group in groups]
        return group_name in group_names

    def _user_account_groups(self, user_id: str) -> list[Group]:
        # This method is used to find all the account groups that a user is a member of.
        groups: list[Group] = []
        account_groups = self._list_account_groups("id,displayName,externalId,members")
        if not account_groups:
            return groups
        for group in account_groups:
            if not group.members:
                continue
            for member in group.members:
                if member.value == user_id:
                    groups.append(group)
                    break
        return groups

    def _list_account_groups(self, scim_attributes: str) -> list[iam.Group]:
        # TODO: we should avoid using this method, as it's not documented
        # get account-level groups even if they're not (yet) assigned to a workspace
        logger.info(f"Listing account groups with {scim_attributes}...")
        account_groups = []
        raw = self._ws.api_client.do("GET", "/api/2.0/account/scim/v2/Groups", query={"attributes": scim_attributes})
        for resource in raw.get("Resources", []):  # type: ignore[union-attr]
            group = iam.Group.from_dict(resource)
            if group.display_name in SYSTEM_GROUPS:
                continue
            account_groups.append(group)
        logger.info(f"Found {len(account_groups)} account groups")
        sorted_groups: list[iam.Group] = sorted(
            account_groups, key=lambda _: _.display_name if _.display_name else ""
        )  # type: ignore[arg-type,return-value]
        return sorted_groups

    def get_mapping(self) -> dict[str, Group]:
        groups: dict[str, Group] = {}
        for group in self._list_account_groups("id,displayName,externalId"):
            if not group.display_name:
                logger.debug(f"Ignoring account group in without name: {group.id}")
                continue
            groups[group.display_name] = group
        return groups


class ConfigureGroups:
    renamed_group_prefix = "db-temp-"
    workspace_group_regex = None
    workspace_group_replace = None
    account_group_regex = None
    group_match_by_external_id = None
    include_group_names = None

    _valid_substitute_pattern = re.compile(r"[\s#,+ \\<>;]")

    def __init__(self, prompts: Prompts):
        self._prompts = prompts
        self._ask_for_group = functools.partial(self._prompts.question, validate=self._is_valid_group_str)
        self._ask_for_substitute = functools.partial(self._prompts.question, validate=self._is_valid_substitute_str)
        self._ask_for_regex = functools.partial(self._prompts.question, validate=self._validate_regex)

    def run(self):
        self.renamed_group_prefix = self._ask_for_group("Backup prefix", default=self.renamed_group_prefix)
        strategy = self._prompts.choice_from_dict(
            "Choose how to map the workspace groups:",
            {
                "Match by Name": lambda: True,
                "Apply a Prefix": self._configure_prefix,
                "Apply a Suffix": self._configure_suffix,
                "Match by External ID": self._configure_external,
                "Regex Substitution": self._configure_substitution,
                "Regex Matching": self._configure_matching,
            },
            sort=False,
        )
        strategy()
        self._configure_names()

    def _configure_prefix(self):
        prefix = self._ask_for_group("Enter a prefix to add to the workspace group name")
        if not prefix:
            return False
        self.workspace_group_regex = "^"
        self.workspace_group_replace = prefix
        return True

    def _configure_suffix(self):
        suffix = self._ask_for_group("Enter a suffix to add to the workspace group name")
        if not suffix:
            return False
        self.workspace_group_regex = "$"
        self.workspace_group_replace = suffix
        return True

    def _configure_substitution(self):
        match_value = self._ask_for_regex("Enter a regular expression for substitution")
        if not match_value:
            return False
        substitute = self._ask_for_substitute("Enter the substitution value")
        if substitute is None:
            return False
        self.workspace_group_regex = match_value
        self.workspace_group_replace = substitute
        return True

    def _configure_matching(self):
        ws_match_value = self._ask_for_regex("Enter a regular expression to match on the workspace group")
        if not ws_match_value:
            return False
        acct_match_value = self._ask_for_regex("Enter a regular expression to match on the account group")
        if not acct_match_value:
            return False
        self.workspace_group_regex = ws_match_value
        self.account_group_regex = acct_match_value
        return True

    def _configure_names(self):
        selected_groups = self._prompts.question(
            "Comma-separated list of workspace group names to migrate. If not specified, we'll use all "
            "account-level groups with matching names to workspace-level groups",
            default="<ALL>",
        )
        if selected_groups != "<ALL>":
            self.include_group_names = [x.strip() for x in selected_groups.split(",")]
        return True

    def _configure_external(self):
        self.group_match_by_external_id = True
        return True

    def _is_valid_group_str(self, group_str: str) -> bool:
        return len(group_str) > 0 and self._is_valid_substitute_str(group_str)

    def _is_valid_substitute_str(self, substitute: str) -> bool:
        return not self._valid_substitute_pattern.search(substitute)

    @staticmethod
    def _validate_regex(regex_input: str) -> bool:
        try:
            re.compile(regex_input)
            return True
        except re.error:
            logger.error(f"{regex_input} is an invalid regular expression")
            return False



from dataclasses import replace
from io import BytesIO
import json
import webbrowser
from pathlib import Path
from configparser import ParsingError

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.installation import Installation, SerdeError
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import AccountClient, WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.workspace import ExportFormat
from databricks.labs.ucx.__about__ import __version__

from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.contexts.account_cli import AccountContext
from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext, LocalCheckoutContext
from databricks.labs.ucx.hive_metastore.tables import What
from databricks.labs.ucx.install import AccountInstaller
from databricks.labs.ucx.source_code.linters.files import LocalCodeLinter
from databricks.labs.ucx.workspace_access.groups import AccountGroupLookup

ucx = App(__file__)
logger = get_logger(__file__)

CANT_FIND_UCX_MSG = (
    "Couldn't find UCX configuration in the user's home folder. "
    "Make sure the current user has configured and installed UCX."
)


def _get_workspace_contexts(
    w: WorkspaceClient, a: AccountClient | None = None, run_as_collection: bool = False, **named_parameters
) -> list[WorkspaceContext]:
    """Get workspace contexts to the workspaces for which the user has access"""
    if not a:
        try:
            a = AccountClient(product='ucx', product_version=__version__)
        except ParsingError as e:
            logger.error("Could not create account client", exc_info=e)
            return []
    account_installer = AccountInstaller(a)
    workspace_contexts = account_installer.get_workspace_contexts(w, run_as_collection, **named_parameters)
    return workspace_contexts


@ucx.command
def workflows(w: WorkspaceClient):
    """Show deployed workflows and their state"""
    ctx = WorkspaceContext(w)
    logger.info("Fetching deployed jobs...")
    latest_job_status = ctx.deployed_workflows.latest_job_status()
    print(json.dumps(latest_job_status))


@ucx.command
def logs(w: WorkspaceClient, workflow: str | None = None):
    """Show logs of the latest job run"""
    ctx = WorkspaceContext(w)
    ctx.deployed_workflows.relay_logs(workflow)


@ucx.command
def open_remote_config(w: WorkspaceClient):
    """Opens remote configuration in the browser"""
    ctx = WorkspaceContext(w)
    workspace_link = ctx.installation.workspace_link('config.yml')
    webbrowser.open(workspace_link)


@ucx.command
def installations(w: WorkspaceClient):
    """Show installations by different users on the same workspace"""
    logger.info("Fetching installations...")
    all_users = []
    for installation in Installation.existing(w, 'ucx'):
        try:
            config = installation.load(WorkspaceConfig)
            all_users.append(
                {
                    'database': config.inventory_database,
                    'path': installation.install_folder(),
                    'warehouse_id': config.warehouse_id,
                }
            )
        except NotFound:
            continue
        except SerdeError:
            continue
    print(json.dumps(all_users))


@ucx.command
def skip(w: WorkspaceClient, schema: str | None = None, table: str | None = None, view: str | None = None) -> None:
    """Create a skip comment on a schema, table or a view."""
    logger.info("Running skip command")
    if not schema:
        logger.error("--schema is a required parameter.")
        return None
    if table and view:
        logger.error("specify --table OR --view, not both")
        return None
    ctx = WorkspaceContext(w)
    table_or_view = table or view
    if table_or_view:
        return ctx.table_mapping.skip_table_or_view(schema, table_or_view, ctx.tables_crawler.load_one)
    return ctx.table_mapping.skip_schema(schema)


@ucx.command
def unskip(w: WorkspaceClient, schema: str | None = None, table: str | None = None, view: str | None = None) -> None:
    """Unset the skip mark from a schema, table or a view."""
    logger.info("Running unskip command")
    if not schema:
        logger.error("--schema is a required parameter.")
        return None
    if table and view:
        logger.error("specify --table OR --view, not both")
        return None
    ctx = WorkspaceContext(w)
    table_or_view = table or view
    if table_or_view:
        return ctx.table_mapping.unskip_table_or_view(schema, table_or_view, ctx.tables_crawler.load_one)
    return ctx.table_mapping.unskip_schema(schema)


@ucx.command(is_account=True)
def sync_workspace_info(a: AccountClient):
    """upload workspace config to all workspaces in the account where ucx is installed"""
    logger.info(f"Account ID: {a.config.account_id}")
    ctx = AccountContext(a)
    ctx.account_workspaces.sync_workspace_info()


@ucx.command(is_account=True)
def report_account_compatibility(a: AccountClient, ctx: AccountContext | None = None, **named_parameters):
    """Report compatibility of all workspaces available in the account"""
    if not ctx:
        ctx = AccountContext(a, named_parameters)
    ctx.account_aggregate.readiness_report()


@ucx.command(is_account=True)
def validate_table_locations(a: AccountClient, ctx: AccountContext | None = None, **named_parameters):
    """Validate if the table locations are overlapping in a workspace and across workspaces"""
    if not ctx:
        ctx = AccountContext(a, named_parameters)
    ctx.account_aggregate.validate_table_locations()


@ucx.command(is_account=True)
def create_account_groups(
    a: AccountClient,
    prompts: Prompts,
    ctx: AccountContext | None = None,
    **named_parameters,
):
    """
    Crawl all workspaces configured in workspace_ids, then creates account level groups if a WS local group is not present
    in the account.
    If workspace_ids is not specified, it will create account groups for all workspaces configured in the account.

    The following scenarios are supported, if a group X:
    - Exist in workspaces A,B,C, and it has same members in there, it will be created in the account
    - Exist in workspaces A,B but not in C, it will be created in the account
    - Exist in workspaces A,B,C. It has same members in A,B, but not in C. Then, X and C_X will be created in the
    account
    """
    if not ctx:
        ctx = AccountContext(a, named_parameters)
    ctx.account_workspaces.create_account_level_groups(prompts)


@ucx.command
def manual_workspace_info(w: WorkspaceClient, prompts: Prompts):
    """only supposed to be run if cannot get admins to run `databricks labs ucx sync-workspace-info`"""
    ctx = WorkspaceContext(w)
    ctx.workspace_info.manual_workspace_info(prompts)


@ucx.command
def create_table_mapping(
    w: WorkspaceClient,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
):
    """create initial table mapping for review"""
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    if ctx:
        workspace_contexts = [ctx]
    for workspace_ctx in workspace_contexts:
        logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
        path = workspace_ctx.table_mapping.save(workspace_ctx.tables_crawler, workspace_ctx.workspace_info)
        if len(workspace_contexts) == 1:
            webbrowser.open(f"{w.config.host}/#workspace{path}")


@ucx.command
def validate_external_locations(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
):
    """validates and provides mapping to external table to external location and shared generation tf scripts"""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for workspace_context in workspace_contexts:
        path = workspace_context.external_locations.save_as_terraform_definitions_on_workspace(
            workspace_context.installation
        )
        if path and prompts.confirm(f"external_locations.tf file written to {path}. Do you want to open it?"):
            webbrowser.open(f"{w.config.host}/#workspace{path}")


@ucx.command
def ensure_assessment_run(w: WorkspaceClient, run_as_collection: bool = False, a: AccountClient | None = None):
    """ensure the assessment job was run on a workspace"""
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for ctx in workspace_contexts:
        workspace_id = ctx.workspace_client.get_workspace_id()
        logger.info(f"Checking assessment workflow in workspace: {workspace_id}")
        deployed_workflows = ctx.deployed_workflows
        # Note: will block if the workflow is already underway but not completed.
        if deployed_workflows.validate_step("assessment"):
            logger.info(f"The assessment workflow has successfully completed in workspace: {workspace_id}")
        else:
            logger.info(f"Starting assessment workflow in workspace: {workspace_id}")
            # If running for a collection, don't wait for each assessment job to finish as that will take a long time.
            deployed_workflows.run_workflow("assessment", skip_job_wait=run_as_collection)


@ucx.command
def update_migration_progress(
    w: WorkspaceClient,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """Manually trigger the migration-progress-experimental job."""
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for ctx in workspace_contexts:
        workspace_id = ctx.workspace_client.get_workspace_id()
        logger.info(f"Starting 'migration-progress-experimental' workflow in workspace: {workspace_id}")
        deployed_workflows = ctx.deployed_workflows
        # If running for a collection, don't wait for each migration-progress job to finish as that will take long time.
        deployed_workflows.run_workflow("migration-progress-experimental", skip_job_wait=run_as_collection)


@ucx.command
def repair_run(w: WorkspaceClient, step):
    """Repair Run the Failed Job"""
    if not step:
        raise KeyError("You did not specify --step")
    ctx = WorkspaceContext(w)
    logger.info(f"Repair Running {step} Job")
    ctx.deployed_workflows.repair_run(step)


@ucx.command
def validate_groups_membership(
    w: WorkspaceClient,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """Validate the groups to see if the groups at account level and workspace level has different membership"""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for workspace_context in workspace_contexts:
        mismatch_groups = workspace_context.group_manager.validate_group_membership()
        print(json.dumps(mismatch_groups))


@ucx.command
def revert_migrated_tables(
    w: WorkspaceClient,
    prompts: Prompts,
    schema: str | None = None,
    table: str | None = None,
    *,
    delete_managed: bool = False,
    ctx: WorkspaceContext | None = None,
):
    """remove notation on a migrated table for re-migration"""
    if not schema and not table:
        question = "You haven't specified a schema or a table. All migrated tables will be reverted. Continue?"
        if not prompts.confirm(question, max_attempts=2):
            return
    if not ctx:
        ctx = WorkspaceContext(w)
    revert = ctx.tables_migrator.print_revert_report(schema=schema, table=table, delete_managed=delete_managed)
    if revert and prompts.confirm("Would you like to continue?", max_attempts=2):
        ctx.tables_migrator.revert_migrated_tables(schema=schema, table=table, delete_managed=delete_managed)


@ucx.command
def move(
    w: WorkspaceClient,
    prompts: Prompts,
    from_catalog: str,
    from_schema: str,
    from_table: str,
    to_catalog: str,
    to_schema: str,
):
    """move a uc table/tables from one schema to another schema in same or different catalog"""
    logger.info("Running move command")
    if from_catalog == "" or to_catalog == "":
        logger.error("Please enter from_catalog and to_catalog details")
        return
    if from_schema == "" or to_schema == "" or from_table == "":
        logger.error("Please enter from_schema, to_schema and from_table (enter * for migrating all tables) details.")
        return
    if from_catalog == to_catalog and from_schema == to_schema:
        logger.error("please select a different schema or catalog to migrate to")
        return
    if not prompts.confirm(f"[WARNING] External tables will be dropped and recreated in the target schema {to_schema}"):
        return
    del_table = prompts.confirm(
        f"should we delete managed tables & views after moving to the new schema" f" {to_catalog}.{to_schema}"
    )
    logger.info(f"migrating tables {from_table} from {from_catalog}.{from_schema} to {to_catalog}.{to_schema}")
    ctx = WorkspaceContext(w)
    ctx.table_move.move(from_catalog, from_schema, from_table, to_catalog, to_schema, del_table=del_table)


@ucx.command
def alias(
    w: WorkspaceClient,
    from_catalog: str,
    from_schema: str,
    from_table: str,
    to_catalog: str,
    to_schema: str,
):
    """move a uc table/tables from one schema to another schema in same or different catalog"""
    if from_catalog == "" or to_catalog == "":
        logger.error("Please enter from_catalog and to_catalog details")
        return
    if from_schema == "" or to_schema == "" or from_table == "":
        logger.error("Please enter from_schema, to_schema and from_table (enter * for migrating all tables) details.")
        return
    if from_catalog == to_catalog and from_schema == to_schema:
        logger.error("please select a different schema or catalog to migrate to")
        return
    logger.info(f"aliasing table {from_table} from {from_catalog}.{from_schema} to {to_catalog}.{to_schema}")
    ctx = WorkspaceContext(w)
    ctx.table_move.alias_tables(from_catalog, from_schema, from_table, to_catalog, to_schema)


@ucx.command
def create_uber_principal(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    **named_parameters,
):
    """For azure cloud, creates a service principal and gives STORAGE BLOB READER access on all the storage account
    used by tables in the workspace and stores the spn info in the UCX cluster policy. For aws,
    it identifies all s3 buckets used by the Instance Profiles configured in the workspace.
    Pass subscription ids for Azure and aws_profile for AWS."""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    for workspace_context in workspace_contexts:
        if workspace_context.is_azure:
            workspace_context.azure_resource_permissions.create_uber_principal(prompts)
        elif workspace_context.is_aws:
            workspace_context.aws_resource_permissions.create_uber_principal(prompts)
        else:
            raise ValueError("Unsupported cloud provider")


@ucx.command
def principal_prefix_access(
    w: WorkspaceClient,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    **named_parameters,
):
    """For azure cloud, identifies all storage accounts used by tables in the workspace, identify spn and its
    permission on each storage accounts. For aws, identifies all the Instance Profiles configured in the workspace and
    its access to all the S3 buckets, along with AWS roles that are set with UC access and its access to S3 buckets.
    The output is stored in the workspace install folder.
    Pass subscription ids for Azure and aws_profile for AWS."""
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    if ctx:
        workspace_contexts = [ctx]
    if w.config.is_azure:
        for workspace_ctx in workspace_contexts:
            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
            workspace_ctx.azure_resource_permissions.save_spn_permissions()
        return
    if w.config.is_aws:
        for workspace_ctx in workspace_contexts:
            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
            instance_role_path = workspace_ctx.aws_resource_permissions.save_instance_profile_permissions()
            logger.info(f"Instance profile and bucket info saved {instance_role_path}")
            logger.info("Generating UC roles and bucket permission info")
            workspace_ctx.aws_resource_permissions.save_uc_compatible_roles()
        return
    raise ValueError("Unsupported cloud provider")


@ucx.command
def create_missing_principals(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    single_role: bool = False,
    role_name="UC_ROLE",
    policy_name="UC_POLICY",
    **named_parameters,
):
    """Not supported for Azure.
    For AWS, this command identifies all the S3 locations that are missing a UC compatible role and creates them.
    By default, it will create a  role per S3 location. Set the optional single_role parameter to True to create a single role for all S3 locations.
    """
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    if ctx:
        workspace_contexts = [ctx]
    if w.config.is_aws:
        for workspace_ctx in workspace_contexts:
            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
            workspace_ctx.iam_role_creation.run(
                prompts, single_role=single_role, role_name=role_name, policy_name=policy_name
            )
    else:
        raise ValueError("Unsupported cloud provider")


@ucx.command
def delete_missing_principals(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    **named_parameters,
):
    """Not supported for Azure.
    For AWS, this command identifies all the UC roles that are created through the create-missing-principals cmd.
    It lists all the UC roles in aws and lets users select the roles to delete. It also validates if the selected roles
    are used by any storage credential and prompts to confirm if roles should still be deleted.
    """
    if not ctx:
        ctx = WorkspaceContext(w, named_parameters)
    if ctx.is_aws:
        return ctx.iam_role_creation.delete_uc_roles(prompts)
    raise ValueError("Unsupported cloud provider")


@ucx.command
def migrate_credentials(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    **named_parameters,
):
    """For Azure, this command prompts to i) create UC storage credentials for the access connectors with a
    managed identity created for each storage account present in the ADLS Gen2 locations, the access connectors are
    granted Storage Blob Data Contributor permissions on their corresponding storage account, to prepare for adopting to
    use Databricks' best practice for using access connectors to authenticate with external storage, and ii) to migrate
    Azure Service Principals, which have Storage Blob Data Contributor, Storage Blob Data Reader, Storage Blob Data
    Owner roles on ADLS Gen2 locations that are being used in Databricks, to UC storage credentials. The Azure Service
    Principals to location mapping are listed in {install_folder}/.ucx/azure_storage_account_info.csv which is generated
    by principal_prefix_access command. Please review the file and delete the Service Principals you do not want to be
    migrated. The command will only migrate the Service Principals that have client secret stored in Databricks Secret.
    For AWS, this command migrates AWS UC compatible roles that are required by Databricks, to UC storage credentials.
    The AWS Roles to location mapping are listed in
    {install_folder}/.ucx/uc_roles_access.csv which is generated by principal_prefix_access command.
    Please review the file and delete the Roles you do not want to be migrated.
    Pass aws_profile for aws.
    """
    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    if ctx:
        workspace_contexts = [ctx]
    if w.config.is_azure:
        for workspace_ctx in workspace_contexts:
            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
            workspace_ctx.service_principal_migration.run(prompts)
    elif w.config.is_aws:
        for workspace_ctx in workspace_contexts:
            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
            workspace_ctx.iam_role_migration.run(prompts)
    else:
        raise ValueError("Unsupported cloud provider")


@ucx.command
def migrate_locations(
    w: WorkspaceClient,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    **named_parameters,
):
    """This command creates UC external locations. The candidate locations to be created are extracted from
    guess_external_locations task in the assessment job. You can run validate_external_locations command to check
    the candidate locations. Please make sure the credentials haven migrated before running this command. The command
    will only create the locations that have corresponded UC Storage Credentials.
    """
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    for workspace_context in workspace_contexts:
        if workspace_context.is_azure or workspace_context.is_aws:
            workspace_context.external_locations_migration.run()
        else:
            raise ValueError("Unsupported cloud provider")


@ucx.command
def create_catalogs_schemas(
    w: WorkspaceClient,
    prompts: Prompts,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """Create UC catalogs and schemas based on the destinations created from create_table_mapping command."""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for workspace_context in workspace_contexts:
        workspace_context.catalog_schema.create_all_catalogs_schemas(prompts)


@ucx.command
def cluster_remap(w: WorkspaceClient, prompts: Prompts):
    """Re-mapping the cluster to UC"""
    logger.info("Remapping the Clusters to UC")
    ctx = WorkspaceContext(w)
    cluster_list = ctx.cluster_access.list_cluster()
    if not cluster_list:
        logger.info("No cluster information present in the workspace")
        return
    print(f"{'Cluster Name':<50}\t{'Cluster Id':<50}")
    for cluster_details in cluster_list:
        print(f"{cluster_details.cluster_name:<50}\t{cluster_details.cluster_id:<50}")
    cluster_ids = prompts.question(
        "Please provide the cluster id's as comma separated value from the above list", default="<ALL>"
    )
    ctx.cluster_access.map_cluster_to_uc(cluster_ids, cluster_list)


@ucx.command
def revert_cluster_remap(w: WorkspaceClient, prompts: Prompts):
    """Reverting Re-mapping of  clusters from UC"""
    logger.info("Reverting the Remapping of the Clusters from UC")
    ctx = WorkspaceContext(w)
    cluster_ids = []
    for cluster_files in ctx.installation.files():
        if cluster_files.path is None:
            continue
        if cluster_files.path.find("backup/clusters") == 0:
            continue
        cluster_id = cluster_files.path.split("/")[-1].split(".")[0]
        cluster_ids.append(cluster_id)
    if not cluster_ids:
        logger.info("There is no cluster files in the backup folder. Skipping the reverting process")
        return
    for cluster in cluster_ids:
        logger.info(cluster)
    cluster_list = prompts.question(
        "Please provide the cluster id's as comma separated value from the above list", default="<ALL>"
    )
    ctx.cluster_access.revert_cluster_remap(cluster_list, cluster_ids)


@ucx.command
def migrate_local_code(w: WorkspaceClient, prompts: Prompts):
    """Fix the code files based on their language."""
    ctx = LocalCheckoutContext(w)
    working_directory = Path.cwd()
    if not prompts.confirm("Do you want to apply UC migration to all files in the current directory?"):
        return
    ctx.local_file_migrator.apply(working_directory)


@ucx.command(is_account=True)
def show_all_metastores(a: AccountClient, workspace_id: str | None = None):
    """Show all metastores in the account"""
    logger.info(f"Account ID: {a.config.account_id}")
    ctx = AccountContext(a)
    ctx.account_metastores.show_all_metastores(workspace_id)


@ucx.command(is_account=True)
def assign_metastore(
    a: AccountClient,
    workspace_id: str | None = None,
    metastore_id: str | None = None,
    default_catalog: str | None = None,
    ctx: AccountContext | None = None,
):
    """Assign metastore to a workspace"""
    if workspace_id is None:
        logger.error("--workspace-id is a required parameter.")
        return
    try:
        workspace_id_casted = int(workspace_id)
    except ValueError:
        logger.error("--workspace-id should be an integer.")
        return
    logger.info(f"Account ID: {a.config.account_id}")
    ctx = ctx or AccountContext(a)
    ctx.account_metastores.assign_metastore(
        ctx.prompts,
        workspace_id_casted,
        metastore_id=metastore_id,
        default_catalog=default_catalog,
    )


@ucx.command
def create_ucx_catalog(w: WorkspaceClient, prompts: Prompts, ctx: WorkspaceContext | None = None) -> None:
    """Create and setup UCX artifact catalog

    Amongst other things, the artifacts are used for tracking the migration progress across workspaces.
    """
    workspace_context = ctx or WorkspaceContext(w)
    workspace_context.catalog_schema.create_ucx_catalog(prompts)
    workspace_context.progress_tracking_installation.run()
    workspace_context.verify_progress_tracking.verify()


@ucx.command
def assign_owner_group(
    w: WorkspaceClient,
    prompts: Prompts,
    *,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """
    Pick owner group. This group will be assigned as owner to all the migrated tables.
    """
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)

    owner_group = AccountGroupLookup(workspace_contexts[0].workspace_client).pick_owner_group(prompts)
    if not owner_group:
        return
    for workspace_context in workspace_contexts:
        config = workspace_context.installation.load(WorkspaceConfig)
        config = replace(config, default_owner_group=owner_group)
        workspace_context.installation.save(config)


@ucx.command
def migrate_tables(
    w: WorkspaceClient,
    prompts: Prompts,
    *,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """
    Trigger the migrate-tables workflow and, optionally, the migrate-external-hiveserde-tables-in-place-experimental
    workflow and migrate-external-tables-ctas.
    """
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for workspace_context in workspace_contexts:
        deployed_workflows = workspace_context.deployed_workflows

        workspace_context.verify_progress_tracking.verify()
        deployed_workflows.run_workflow("migrate-tables")

        tables = list(workspace_context.tables_crawler.snapshot())
        hiveserde_tables = [table for table in tables if table.what == What.EXTERNAL_HIVESERDE]
        if len(hiveserde_tables) > 0:
            percentage_hiveserde_tables = len(hiveserde_tables) / len(tables) * 100
            if prompts.confirm(
                f"Found {len(hiveserde_tables)} ({percentage_hiveserde_tables:.2f}%) hiveserde tables in "
                f"{workspace_context.workspace_client.config.host}, do you want to run "
                f"the `migrate-external-hiveserde-tables-in-place-experimental` workflow?"
            ):
                deployed_workflows.run_workflow("migrate-external-hiveserde-tables-in-place-experimental")

        external_ctas_tables = [table for table in tables if table.what == What.EXTERNAL_NO_SYNC]
        if len(external_ctas_tables) > 0:
            percentage_external_ctas_tables = len(external_ctas_tables) / len(tables) * 100
            if prompts.confirm(
                f"Found {len(external_ctas_tables)} ({percentage_external_ctas_tables:.2f}%) external tables which "
                f"cannot be migrated using sync in {workspace_context.workspace_client.config.host}, do you want to "
                "run the `migrate-external-tables-ctas` workflow?"
            ):
                deployed_workflows.run_workflow("migrate-external-tables-ctas")


@ucx.command
def migrate_acls(
    w: WorkspaceClient,
    *,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
    **named_parameters,
):
    """
    Migrate the ACLs for migrated tables and view. Can work with hms federation or other table migration scenarios.
    """
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
    target_catalog = named_parameters.get("target_catalog")
    hms_fed = named_parameters.get("hms_fed", False)
    dry_run = named_parameters.get("dry_run", False)
    if dry_run:
        total_grants = 0
        for workspace_context in workspace_contexts:
            grants = workspace_context.acl_migrator.snapshot()
            total_grants += len(grants)
        logger.info(
            f"Dry run completed. Found {total_grants} grants. The crawled grants can be found in the 'inferred_grants' table. "
            "No changes were made."
        )
        urls: str = ""
        for workspace_context in workspace_contexts:
            urls += (
                f"{workspace_context.connect_config.host}/explore/data/hive_metastore/"
                f"{workspace_context.config.inventory_database}/inferred_grants\n"
            )
        logger.info(f"URLs to the inferred grants tables: \n{urls}")
        return
    for workspace_context in workspace_contexts:
        workspace_context.acl_migrator.migrate_acls(target_catalog=target_catalog, hms_fed=hms_fed)


@ucx.command
def migrate_dbsql_dashboards(
    w: WorkspaceClient,
    dashboard_id: str | None = None,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """Migrate table references in DBSQL Dashboard queries"""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
    for workspace_context in workspace_contexts:
        if dashboard_id:
            workspace_context.redash.migrate_dashboards(dashboard_id)
        else:
            workspace_context.redash.migrate_dashboards()


@ucx.command
def revert_dbsql_dashboards(w: WorkspaceClient, dashboard_id: str | None = None, ctx: WorkspaceContext | None = None):
    """Revert migrated DBSQL Dashboard queries back to their original state"""
    ctx = ctx or WorkspaceContext(w)
    ctx.redash_crawler.snapshot(force_refresh=True)  # Need the latest tags before reverting dashboards
    if dashboard_id:
        ctx.redash.revert_dashboards(dashboard_id)
    else:
        ctx.redash.revert_dashboards()


@ucx.command(is_account=True)
def join_collection(a: AccountClient, workspace_ids: str):
    """joins the workspace to an existing collection"""
    account_installer = AccountInstaller(a)
    w_ids = [int(_.strip()) for _ in workspace_ids.split(",") if _]
    account_installer.join_collection(w_ids)


@ucx.command
def upload(
    file: Path | str,
    w: WorkspaceClient,
    run_as_collection: bool = False,
    a: AccountClient | None = None,  # Only used while testing
):
    """Upload a file to the (collection of) workspace(s)"""
    file = Path(file)
    contexts = _get_workspace_contexts(w, run_as_collection=run_as_collection, a=a)
    logger.warning("The schema of CSV files is NOT validated, ensure it is correct")
    for ctx in contexts:
        ctx.installation.upload(file.name, file.read_bytes())
    if len(contexts) > 0:
        logger.info(f"Finished uploading {file}")


@ucx.command
def download(
    file: Path | str,
    w: WorkspaceClient,
    run_as_collection: bool = False,
    a: AccountClient | None = None,  # Only used while testing
):
    """Download and merge a CSV file from the ucx installation in a (collection of) workspace(s)"""
    file = Path(file)
    if file.suffix != ".csv":
        raise ValueError("Command only supported for CSV files")
    contexts = _get_workspace_contexts(w, run_as_collection=run_as_collection, a=a)
    csv_header = None
    with file.open("wb") as output:
        for ctx in contexts:
            remote_file_name = f"{ctx.installation.install_folder()}/{file.name}"
            try:
                # Installation does not have a download method
                data = ctx.workspace_client.workspace.download(remote_file_name, format=ExportFormat.AUTO).read()
            except NotFound:
                logger.warning(f"File not found for {ctx.workspace_client.config.host}: {remote_file_name}")
                continue
            input_ = BytesIO()  # BytesIO supports .readline() to read the header, where StreamingResponse does not
            input_.write(data.rstrip(b"\n"))
            input_.seek(0)  # Go back to the beginning of the file
            csv_header_next = input_.readline()
            if csv_header is None:
                csv_header = csv_header_next
                output.write(csv_header)
            elif csv_header == csv_header_next:
                output.write(b"\n")
            else:
                raise ValueError("CSV files have different headers")
            output.write(input_.read())
    if csv_header is None:
        logger.warning("No file(s) to download found")
    if file.is_file() and file.stat().st_size == 0:
        file.unlink()
    else:
        logger.info(f"Finished downloading {file}")


@ucx.command
def lint_local_code(
    w: WorkspaceClient, prompts: Prompts, path: str | None = None, ctx: LocalCheckoutContext | None = None
):
    """Lint local code files looking for problems."""
    if ctx is None:
        ctx = LocalCheckoutContext(w)
    linter: LocalCodeLinter = ctx.local_code_linter
    linter.lint(prompts, None if path is None else Path(path))


@ucx.command
def export_assessment(w: WorkspaceClient, prompts: Prompts):
    """Export the UCX assessment queries to a zip file."""
    ctx = WorkspaceContext(w)
    exporter = ctx.assessment_exporter
    exporter.export_results(prompts)


@ucx.command
def create_federated_catalog(w: WorkspaceClient, prompts: Prompts):
    """(Experimental) Create federated catalog from current workspace Hive Metastore."""
    ctx = WorkspaceContext(w)
    ctx.federation.create_from_cli(prompts)


@ucx.command
def enable_hms_federation(w: WorkspaceClient, _: Prompts, ctx: WorkspaceContext | None = None):
    """(Experimental) Create federated catalog from current workspace Hive Metastore."""
    if not ctx:
        ctx = WorkspaceContext(w)
    ctx.federation_enabler.enable()


@ucx.command
def migrate_dlt_pipelines(
    w: WorkspaceClient,
    ctx: WorkspaceContext | None = None,
    run_as_collection: bool = False,
    a: AccountClient | None = None,
) -> None:
    """Migrate DLT pipelines to UC"""
    if ctx:
        workspace_contexts = [ctx]
    else:
        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)

    for workspace_context in workspace_contexts:
        workspace_context.pipelines_migrator.migrate_pipelines()


if __name__ == "__main__":
    ucx()



import collections
import logging
from collections.abc import Iterable
from dataclasses import replace

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.assessment.jobs import JobInfo, JobOwnership
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.progress.history import ProgressEncoder
from databricks.labs.ucx.progress.install import Historical
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler
from databricks.labs.ucx.source_code.jobs import JobProblem


logger = logging.getLogger(__name__)
JobIdToFailuresType = dict[str, list[str]]  # dict[<job id>, list[<failure message>]]


class JobsProgressEncoder(ProgressEncoder[JobInfo]):
    """Encoder class:Job to class:History."""

    def __init__(
        self,
        sql_backend: SqlBackend,
        ownership: JobOwnership,
        direct_fs_access_crawlers: list[DirectFsAccessCrawler],
        inventory_database: str,
        run_id: int,
        workspace_id: int,
        catalog: str,
    ) -> None:
        super().__init__(
            sql_backend,
            ownership,
            JobInfo,
            run_id,
            workspace_id,
            catalog,
            "multiworkspace",
            "historical",
        )
        self._direct_fs_access_crawlers = direct_fs_access_crawlers
        self._inventory_database = inventory_database

    def append_inventory_snapshot(self, snapshot: Iterable[JobInfo]) -> None:
        job_problems = self._get_job_problems()
        dfsas = self._get_direct_filesystem_accesses()
        history_records = []
        for record in snapshot:
            history_record = self._encode_job_as_historical(record, job_problems, dfsas)
            history_records.append(history_record)
        logger.debug(f"Appending {len(history_records)} {self._klass} table record(s) to history.")
        # The mode is 'append'. This is documented as conflict-free.
        self._sql_backend.save_table(escape_sql_identifier(self.full_name), history_records, Historical, mode="append")

    def _get_job_problems(self) -> JobIdToFailuresType:
        index = collections.defaultdict(list)
        for row in self._sql_backend.fetch(
            'SELECT * FROM workflow_problems',
            catalog='hive_metastore',
            schema=self._inventory_database,
        ):
            job_problem = JobProblem(**row.asDict())
            failure = f'{job_problem.code}: {job_problem.task_key} task: {job_problem.path}: {job_problem.message}'
            index[str(job_problem.job_id)].append(failure)
        return index

    def _get_direct_filesystem_accesses(self) -> JobIdToFailuresType:
        index = collections.defaultdict(list)
        for crawler in self._direct_fs_access_crawlers:
            for direct_fs_access in crawler.snapshot():
                # The workflow and task source lineage are added by the WorkflowLinter
                if len(direct_fs_access.source_lineage) < 2:
                    continue
                if direct_fs_access.source_lineage[0].object_type != "WORKFLOW":
                    continue
                if direct_fs_access.source_lineage[1].object_type != "TASK":
                    continue
                job_id = direct_fs_access.source_lineage[0].object_id
                task_key = direct_fs_access.source_lineage[1].object_id  # <job id>/<task key>
                # Follow same failure message structure as the JobProblems above and DirectFsAccessPyLinter deprecation
                code = "direct-filesystem-access"
                message = f"The use of direct filesystem references is deprecated: {direct_fs_access.path}"
                failure = f"{code}: {task_key} task: {direct_fs_access.source_id}: {message}"
                index[job_id].append(failure)
        return index

    def _encode_job_as_historical(
        self,
        record: JobInfo,
        job_problems: JobIdToFailuresType,
        dfsas: JobIdToFailuresType,
    ) -> Historical:
        """Encode a job as a historical records.

        Failures are detected by the WorkflowLinter:
        - Job problems
        - Direct filesystem access by code used in job
        """
        historical = super()._encode_record_as_historical(record)
        failures = []
        failures.extend(job_problems.get(record.job_id, []))
        failures.extend(dfsas.get(record.job_id, []))
        return replace(historical, failures=historical.failures + failures)



import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from datetime import timedelta
from functools import cached_property
from typing import Generic, TypeVar, final

from databricks.labs.blueprint.paths import WorkspacePath
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, InternalError
from databricks.sdk.retries import retried
from databricks.sdk.service.iam import User, PermissionLevel
from databricks.sdk.service.workspace import ObjectType

logger = logging.getLogger(__name__)


Record = TypeVar("Record")


class AdministratorFinder(ABC):
    def __init__(self, ws: WorkspaceClient) -> None:
        self._ws = ws

    @abstractmethod
    def find_admin_users(self) -> Iterable[User]:
        """Locate active admin users."""
        raise NotImplementedError()


class WorkspaceAdministratorFinder(AdministratorFinder):
    """Locate the users that are in the 'admin' workspace group for a given workspace."""

    @staticmethod
    def _member_of_group_named(user: User, group_name: str) -> bool:
        """Determine whether a user belongs to a group with the given name or not."""
        return user.groups is not None and any(g.display == group_name for g in user.groups)

    @staticmethod
    def _member_of_group(user: User, group_id: str) -> bool:
        """Determine whether a user belongs to a group with the given identifier or not."""
        return user.groups is not None and any(g.value == group_id for g in user.groups)

    def _is_active_admin(self, user: User) -> bool:
        """Determine if a user is an active administrator."""
        return bool(user.active) and self._member_of_group_named(user, "admins")

    def _is_workspace_group(self, group_id: str) -> bool:
        """Determine whether a group_id corresponds to a workspace group or not."""
        try:
            group = self._ws.groups.get(group_id)
        except NotFound:
            return False
        return bool(group.meta and group.meta.resource_type == "WorkspaceGroup")

    def find_admin_users(self) -> Iterable[User]:
        """Enumerate the active workspace administrators in a given workspace.

        Returns:
            Iterable[User]: The active workspace administrators, if any.
        """
        logger.debug("Enumerating users to locate active workspace administrators...")
        all_users = self._ws.users.list(attributes="id,active,userName,groups")
        # The groups attribute is a flattened list of groups a user belongs to; hunt for the 'admins' workspace group.
        # Reference: https://learn.microsoft.com/en-us/azure/databricks/admin/users-groups/groups#account-vs-workspace-group
        admin_users = [user for user in all_users if user.user_name and self._is_active_admin(user)]
        logger.debug(f"Verifying membership of the 'admins' workspace group for users: {admin_users}")
        maybe_admins_id = set()
        for user in admin_users:
            if not user.groups:
                continue
            for group in user.groups:
                if group.display == "admins" and group.value:
                    maybe_admins_id.add(group.value)
        # There can only be a single 'admins' workspace group.
        for group_id in maybe_admins_id:
            if self._is_workspace_group(group_id):
                return (user for user in admin_users if self._member_of_group(user, group_id))
        return ()


class AccountAdministratorFinder(AdministratorFinder):
    """Locate the users that are account administrators for this workspace."""

    @staticmethod
    def _has_role(user: User, role: str) -> bool:
        """Determine whether a user has a given role or not."""
        return user.roles is not None and any(r.value == role for r in user.roles)

    def find_admin_users(self) -> Iterable[User]:
        """Enumerate the active account administrators associated with a given workspace.

        Returns:
            Iterable[User]: The active account administrators, if any.
        """
        logger.debug("Enumerating account users to locate active administrators...")
        response = self._ws.api_client.do(
            "GET", "/api/2.0/account/scim/v2/Users", query={"attributes": "id,active,userName,roles"}
        )
        assert isinstance(response, dict)
        all_users = (User.from_dict(resource) for resource in response.get("Resources", []))
        # Reference: https://learn.microsoft.com/en-us/azure/databricks/admin/users-groups/groups#account-admin
        return (user for user in all_users if user.active and user.user_name and self._has_role(user, "account_admin"))


class AdministratorLocator:
    """Locate a workspace administrator, if possible.

    This will first try to find an active workspace administrator. If there are multiple, the first (alphabetically
    sorted by user-name) will be used. If no active workspace administrators can be found then an account administrator
    is sought, again returning the first alphabetically by user-name if more than one is found.
    """

    def __init__(
        self,
        ws: WorkspaceClient,
        *,
        finders: Sequence[Callable[[WorkspaceClient], AdministratorFinder]] = (
            WorkspaceAdministratorFinder,
            AccountAdministratorFinder,
        ),
    ) -> None:
        """
        Initialize the instance, which will try to locate administrators using the workspace for the supplied client.

        Args:
            ws (WorkspaceClient): the client for workspace in which to locate admin users.
            finders: a sequence of factories that will be instantiated on demand to locate admin users.
        """
        self._ws = ws
        self._finders = finders

    @cached_property
    def _workspace_id(self) -> int:
        # Makes a REST call, so we cache it.
        return self._ws.get_workspace_id()

    @cached_property
    def _found_admin(self) -> str | None:

        # Ordering helper: User.user_name is typed as optional but we can't sort by None.
        # (The finders already filter out users without a user-name.)
        def _by_username(user: User) -> str:
            assert user.user_name
            return user.user_name

        # Lazily instantiate and query the finders in an attempt to locate an admin user.
        for factory in self._finders:
            finder = factory(self._ws)
            # First alphabetically by name.
            admin_user = min(finder.find_admin_users(), default=None, key=_by_username)
            if admin_user:
                return admin_user.user_name
        return None

    def get_workspace_administrator(self) -> str:
        """The user-name of an admin user for the workspace.

        Raises:
              RuntimeError if an admin user cannot be found in the current workspace.
        """
        found_admin = self._found_admin
        if found_admin is None:
            msg = f"No active workspace or account administrator can be found for workspace: {self._workspace_id}"
            raise RuntimeError(msg)
        return found_admin


class Ownership(ABC, Generic[Record]):
    """Determine an owner for a given type of object."""

    def __init__(self, administrator_locator: AdministratorLocator) -> None:
        self._administrator_locator = administrator_locator

    @final
    def owner_of(self, record: Record) -> str:
        """Obtain the user-name of a user that is responsible for the given record.

        This is intended to be a point of contact, and is either:

         - A user directly associated with the resource, such as the original creator; or
         - An active administrator for the current workspace.

        Args:
            record (Record): The record for which an associated user-name is sought.
        Returns:
            A string containing the user-name attribute of a user considered to be responsible for the resource.
        Raises:
            RuntimeError if there are no active administrators for the current workspace.
        """
        return self._maybe_direct_owner(record) or self._administrator_locator.get_workspace_administrator()

    @abstractmethod
    def _maybe_direct_owner(self, record: Record) -> str | None:
        """Obtain the record-specific user-name associated with the given record, if any."""
        return None


class WorkspacePathOwnership(Ownership[WorkspacePath]):
    def __init__(self, administrator_locator: AdministratorLocator, ws: WorkspaceClient) -> None:
        super().__init__(administrator_locator)
        self._ws = ws

    def owner_of_path(self, path: str) -> str:
        return self.owner_of(WorkspacePath(self._ws, path))

    @retried(on=[InternalError], timeout=timedelta(minutes=1))
    def _maybe_direct_owner(self, record: WorkspacePath) -> str | None:
        maybe_type_and_id = self._maybe_type_and_id(record)
        if not maybe_type_and_id:
            return None
        object_type, object_id = maybe_type_and_id
        try:
            object_permissions = self._ws.permissions.get(object_type, object_id)
            return self._infer_from_first_can_manage(object_permissions)
        except NotFound:
            logger.warning(f"removed on backend: {object_type} {object_id}")
            return None

    @staticmethod
    def _maybe_type_and_id(path: WorkspacePath) -> tuple[str, str] | None:
        object_info = path._object_info  # pylint: disable=protected-access
        object_id = str(object_info.object_id)
        match object_info.object_type:
            case ObjectType.NOTEBOOK:
                return 'notebooks', object_id
            case ObjectType.FILE:
                return 'files', object_id
        return None

    @staticmethod
    def _infer_from_first_can_manage(object_permissions):
        for acl in object_permissions.access_control_list:
            for permission in acl.all_permissions:
                if permission.permission_level != PermissionLevel.CAN_MANAGE:
                    continue
                if acl.user_name:
                    return acl.user_name
                if acl.group_name:
                    return acl.group_name
                return acl.service_principal_name
        return None


class LegacyQueryOwnership(Ownership[str]):
    def __init__(self, administrator_locator: AdministratorLocator, workspace_client: WorkspaceClient) -> None:
        super().__init__(administrator_locator)
        self._workspace_client = workspace_client

    def _maybe_direct_owner(self, record: str) -> str | None:
        try:
            legacy_query = self._workspace_client.queries.get(record)
            return legacy_query.owner_user_name
        except NotFound:
            return None
        except InternalError:  # redash is very naughty and throws 500s instead of proper 404s
            return None



import logging
from pathlib import Path

from databricks.labs.blueprint.tui import Prompts

from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.dashboards import DashboardMetadata

logger = logging.getLogger(__name__)


class AssessmentExporter:

    def __init__(self, sql_backend: SqlBackend, config: WorkspaceConfig):
        self._sql_backend = sql_backend
        self._config = config

    def export_results(self, prompts: Prompts):
        """Main method to export results to CSV files inside a ZIP archive."""
        project_root = Path(__file__).resolve().parents[3]
        queries_path_root = project_root / "labs/ucx/queries/assessment"

        results_directory = Path(
            prompts.question(
                "Choose a path to save the UCX Assessment results",
                default=Path.cwd().as_posix(),
                validate=lambda p_: Path(p_).exists(),
            )
        )

        query_choice = prompts.choice(
            "Choose which assessment results to export",
            [subdir.name for subdir in queries_path_root.iterdir() if subdir.is_dir()],
        )

        export_path = results_directory / f"export_{query_choice}_results.zip"
        queries_path = queries_path_root / query_choice

        assessment_results = DashboardMetadata.from_path(queries_path).replace_database(
            database=self._config.inventory_database, database_to_replace="inventory"
        )

        logger.info("Exporting assessment results....")
        results_path = assessment_results.export_to_zipped_csv(self._sql_backend, export_path)
        logger.info(f"Results exported to {results_path}")

        return results_path



import logging
from functools import partial
from typing import BinaryIO

from databricks.labs.blueprint.parallel import Threads
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError
from databricks.sdk.service.jobs import PipelineTask, Task, JobSettings

from databricks.labs.ucx.assessment.jobs import JobsCrawler
from databricks.labs.ucx.assessment.pipelines import PipelineInfo, PipelinesCrawler

logger = logging.getLogger(__name__)


class PipelinesMigrator:
    """
    PipelinesMigrator is responsible for migrating pipelines from HMS to UC
    It uses the DLT Migration API to migrate the pipelines and also updates the jobs associated with the pipelines if any
    The class also provides an option to skip the pipelines that are already migrated or the ones that are explicitly skipped

    :param ws: WorkspaceClient
    :param pipelines_crawler: PipelinesCrawler
    :param catalog_name: str
    :param skip_pipeline_ids: list[str] | None
    """

    def __init__(
        self,
        ws: WorkspaceClient,
        pipelines_crawler: PipelinesCrawler,
        jobs_crawler: JobsCrawler,
        catalog_name: str,
        skip_pipeline_ids: list[str] | None = None,
    ):
        self._ws = ws
        self._pipeline_crawler = pipelines_crawler
        self._jobs_crawler = jobs_crawler
        self._catalog_name = catalog_name
        self._skip_pipeline_ids = skip_pipeline_ids or []
        self._pipeline_job_tasks_mapping: dict[str, list[dict]] = {}

    def _populate_pipeline_job_tasks_mapping(self) -> None:
        """
        Populates the pipeline_job_tasks_mapping dictionary with the pipeline_id as key and the list of jobs associated with the pipeline
        """
        jobs = self._jobs_crawler.snapshot()

        for job in jobs:
            if not job.job_id:
                continue

            job_details = self._ws.jobs.get(int(job.job_id))
            if not job_details.settings or not job_details.settings.tasks:
                continue

            for task in job_details.settings.tasks:
                if not task.pipeline_task:
                    continue
                pipeline_id = task.pipeline_task.pipeline_id
                job_info = {"job_id": job.job_id, "task_key": task.task_key}
                if pipeline_id not in self._pipeline_job_tasks_mapping:
                    self._pipeline_job_tasks_mapping[pipeline_id] = [job_info]
                else:
                    self._pipeline_job_tasks_mapping[pipeline_id].append(job_info)
                logger.info(f"Found job:{job.job_id} task:{task.task_key} associated with pipeline {pipeline_id}")

    def _get_pipelines_to_migrate(self) -> list[PipelineInfo]:
        """
        Returns the list of pipelines in the current workspace
        """
        return list(self._pipeline_crawler.snapshot())

    def migrate_pipelines(self) -> None:
        """
        Migrate the pipelines from HMS to UC. Public method to be called to start the pipeline migration process
        """
        self._populate_pipeline_job_tasks_mapping()
        self._migrate_pipelines()

    def _migrate_pipelines(self) -> list[partial[dict | bool | list | BinaryIO]]:
        """
        Create tasks to parallely migrate the pipelines
        """
        # get pipelines to migrate
        pipelines_to_migrate = self._get_pipelines_to_migrate()
        logger.info(f"Found {len(pipelines_to_migrate)} pipelines to migrate")

        tasks = []
        for pipeline in pipelines_to_migrate:
            if pipeline.pipeline_id in self._skip_pipeline_ids:
                logger.info(f"Skipping pipeline {pipeline.pipeline_id}")
                continue
            tasks.append(partial(self._migrate_pipeline, pipeline))
        if not tasks:
            return []
        Threads.strict("migrate pipelines", tasks)
        return tasks

    def _migrate_pipeline(self, pipeline: PipelineInfo) -> dict | list | BinaryIO | bool:
        """
        Private method to clone the pipeline and handle the exceptions
        """
        try:
            return self._clone_pipeline(pipeline)
        except DatabricksError as e:
            if "Cloning from Hive Metastore to Unity Catalog is currently not supported" in str(e):
                logger.error(f"{e}: Please contact Databricks to enable DLT HMS to UC migration API on this workspace")
                return False
            logger.error(f"Failed to migrate pipeline {pipeline.pipeline_id}: {e}")
            return False

    def _clone_pipeline(self, pipeline: PipelineInfo) -> dict | list | BinaryIO:
        """
        This method calls the DLT Migration API to clone the pipeline
        Stop and rename the old pipeline before cloning the new pipeline
        Call the DLT Migration API to clone the pipeline
        Update the jobs associated with the pipeline to point to the new pipeline
        """
        # Need to get the pipeline again to get the libraries
        # else updating name fails with libraries not provided error
        get_pipeline = self._ws.pipelines.get(pipeline.pipeline_id)
        if get_pipeline.spec:
            if get_pipeline.spec.catalog:
                # Skip if the pipeline is already migrated to UC
                logger.info(f"Pipeline {pipeline.pipeline_id} is already migrated to UC")
                return []

            # Stop HMS pipeline
            self._ws.pipelines.stop(pipeline.pipeline_id)
            # Rename old pipeline first
            self._ws.pipelines.update(
                pipeline.pipeline_id,
                name=f"{pipeline.pipeline_name} [OLD]",
                clusters=get_pipeline.spec.clusters if get_pipeline.spec.clusters else None,
                storage=get_pipeline.spec.storage if get_pipeline.spec.storage else None,
                continuous=get_pipeline.spec.continuous if get_pipeline.spec.continuous else None,
                deployment=get_pipeline.spec.deployment if get_pipeline.spec.deployment else None,
                target=get_pipeline.spec.target if get_pipeline.spec.target else None,
                libraries=get_pipeline.spec.libraries if get_pipeline.spec.libraries else None,
            )

        # Clone pipeline
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }
        body = {
            'catalog': self._catalog_name,
            'clone_mode': 'MIGRATE_TO_UC',
            'configuration': {'pipelines.migration.ignoreExplicitPath': 'true'},
            'name': f"{pipeline.pipeline_name}",
        }
        res = self._ws.api_client.do(
            'POST', f'/api/2.0/pipelines/{pipeline.pipeline_id}/clone', body=body, headers=headers
        )
        assert isinstance(res, dict)
        if 'pipeline_id' not in res:
            logger.warning(f"Failed to clone pipeline {pipeline.pipeline_id}")
            return res

        # After successful clone, update jobs
        # Currently there is no SDK method available to migrate the DLT pipelines
        # We are directly using the DLT Migration API in the interim, once the SDK method is available, we can replace this
        if pipeline.pipeline_id in self._pipeline_job_tasks_mapping:
            for pipeline_job_task_mapping in self._pipeline_job_tasks_mapping[pipeline.pipeline_id]:
                self._ws.jobs.update(
                    pipeline_job_task_mapping['job_id'],
                    new_settings=JobSettings(
                        tasks=[
                            Task(
                                pipeline_task=PipelineTask(pipeline_id=str(res.get('pipeline_id'))),
                                task_key=pipeline_job_task_mapping['task_key'],
                            )
                        ]
                    ),
                )
                logger.info(f"Updated job {pipeline_job_task_mapping['job_id']} with new pipeline {res['pipeline_id']}")

        # TODO:
        # Check the error from UI
        # BAD_REQUEST: Standard_D4pds_v6 is a Graviton instance and is not compatible with runtime dlt:14.1.21-delta-pipelines-dlt-release-2024.41-rc0-commit-f32d838-image-894c190.
        return res



import collections
import datetime as dt
import logging
from dataclasses import dataclass
from pathlib import PurePath

from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError, NotFound
from databricks.sdk.retries import retried

from databricks.labs.ucx.hive_metastore.grants import MigrateGrants
from databricks.labs.ucx.hive_metastore.mapping import TableMapping

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Catalog:
    """Represents a catalog from Unity Catalog.

    The Databricks SDK also comes with a representation for a catalog: `databricks.sdk.service.catalog.CatalogInfo`.
    However, we introduce this dataclass to have a minimal, extensible representation required for UCX.

    Docs:
        https://docs.databricks.com/en/data-governance/unity-catalog/manage-privileges/privileges.html#securable-objects-in-unity-catalog
    """

    name: str
    """The catalog name"""

    @property
    def full_name(self) -> str:
        """The full name of the catalog.

        For a catalog, this is same as the attr:name as the catalog is the top of the object hierarchy (see doc link
        above).
        """
        return self.name

    @property
    def key(self) -> str:
        """Synonym for attr:full_name:."""
        return self.full_name

    @property
    def kind(self) -> str:
        """The object kind.

        Note:
            In the SDK this maps to attr:securable_type.

        TODO:
            https://github.com/databrickslabs/ucx/issues/2975
        """
        return "CATALOG"


@dataclass(frozen=True)
class Schema:
    """Represents a schema from Unity Catalog.

    The Databricks SDK also comes with a representation for a schema: `databricks.sdk.service.catalog.SchemaInfo`.
    However, we introduce this dataclass to have a minimal, extensible representation required for UCX.

    Docs:
        https://docs.databricks.com/en/data-governance/unity-catalog/manage-privileges/privileges.html#securable-objects-in-unity-catalog
    """

    catalog: str
    """The catalog the schema is part of.

    Note:
        Maps to `SchemaInfo.catalog_name`, when introducing this class `catalog` is consistent with
        `databricks.labs.ucx.hive_metastore.tables.Table.catalog`.
    """

    name: str
    """The schema name"""

    @property
    def full_name(self) -> str:
        """The full name of the schema.

        For a schema, the second layer of the object hierarchy (see doc link above).
        """
        return f"{self.catalog}.{self.name}"

    @property
    def key(self) -> str:
        """Synonym for attr:full_name."""
        return self.full_name

    @property
    def kind(self) -> str:
        """The object kind.

        Below "DATABASE" is chosen as this is the kind used in the grants module. However, more consistent with
        Databricks documentation would be to use "SCHEMA" instead:
        https://docs.databricks.com/en/data-governance/table-acls/object-privileges.html#securable-objects-in-the-hive-metastore

        TODO:
            https://github.com/databrickslabs/ucx/issues/2975
        """
        return "DATABASE"  # TODO: https://github.com/databrickslabs/ucx/issues/2974


class CatalogSchema:

    def __init__(
        self,
        ws: WorkspaceClient,
        table_mapping: TableMapping,
        migrate_grants: MigrateGrants,
        ucx_catalog: str,
        *,
        timeout: dt.timedelta | None = dt.timedelta(seconds=30),
    ):
        self._ws = ws
        self._table_mapping = table_mapping
        self._migrate_grants = migrate_grants
        self._external_locations = list(self._ws.external_locations.list())
        self._ucx_catalog = ucx_catalog
        self._timeout = timeout

    def create_ucx_catalog(self, prompts: Prompts, *, properties: dict[str, str] | None = None) -> None:
        """Create the UCX catalog.

        Args:
            prompts : Prompts
                The prompts object to use for interactive input.
            properties : (dict[str, str] | None), default None
                The properties to pass to the catalog. If None, no properties are passed.
        """
        self._create_catalog_validate(Catalog(self._ucx_catalog), prompts, properties=properties)

    def create_all_catalogs_schemas(self, prompts: Prompts, *, properties: dict[str, str] | None = None) -> None:
        """Create all UC catalogs and schemas reference by the table mapping file.

        After creation, the grants from the HIVE metastore schemas are applied to the matching UC catalogs and schemas.
        """
        catalogs, schemas = self._catalogs_schemas_from_table_mapping()
        for dst_catalog, src_schemas in catalogs.items():
            self._create_catalog_validate(dst_catalog, prompts, properties=properties)
        for dst_schema, src_schemas in schemas.items():
            self._create_schema(dst_schema)
            for src_schema in src_schemas:
                self._migrate_grants.apply(src_schema, dst_schema)
        # Apply catalog grants as last to avoid transferring ownership before schema grants are applied
        for dst_catalog, src_schemas in catalogs.items():
            for src_schema in src_schemas:
                self._migrate_grants.apply(src_schema, dst_catalog)

    def _catalogs_schemas_from_table_mapping(self) -> tuple[dict[Catalog, set[Schema]], dict[Schema, set[Schema]]]:
        """Generate a list of catalogs and schema to be created from table mapping.

        For applying grants after creating the catalogs and schemas, we track the HIVE metastore schemas from which the
        UC catalog or schema is mapped.

        :returns
            dict[Catalog, set[Schema]] : The UC catalogs to create with the schemas it is mapped from.
            dict[Schema, set[Schema]] : The UC schemas to create with the schemas it is mapped from.
        """
        catalogs, schemas = collections.defaultdict(set), collections.defaultdict(set)
        for mappings in self._table_mapping.load():
            src_schema = Schema("hive_metastore", mappings.src_schema)
            dst_catalog = Catalog(mappings.catalog_name)
            dst_schema = Schema(mappings.catalog_name, mappings.dst_schema)
            catalogs[dst_catalog].add(src_schema)
            schemas[dst_schema].add(src_schema)
        return catalogs, schemas

    def _create_catalog_validate(
        self,
        catalog: Catalog,
        prompts: Prompts,
        *,
        properties: dict[str, str] | None,
    ) -> Catalog:
        catalog_existing = self._get_catalog(catalog)
        if catalog_existing:
            logger.warning(f"Skipping already existing catalog: {catalog.name}")
            return catalog_existing
        logger.info(f"Validating UC catalog: {catalog.name}")
        attempts = 3
        while True:
            catalog_storage = prompts.question(
                f"Please provide storage location url for catalog: {catalog.name}", default="metastore"
            )
            if self._validate_location(catalog_storage):
                break
            attempts -= 1
            if attempts == 0:
                raise NotFound(f"Failed to validate location for catalog: {catalog.name}")
        return self._create_catalog(catalog, catalog_storage, properties=properties)

    def _validate_location(self, location: str) -> bool:
        if location == "metastore":
            return True
        try:
            PurePath(location)
        except ValueError:
            logger.error(f"Invalid location path: {location}")
            return False
        for external_location in self._external_locations:
            if external_location.url is not None and location.startswith(external_location.url):
                return True
        logger.warning(f"No matching external location found for: {location}")
        return False

    def _get_catalog(self, catalog: Catalog) -> Catalog | None:
        """Get a catalog.

        Args:
            catalog (Catalog) : The catalog to get.

        Returns:
            Catalog : The catalog it got.
            None : If the catalog does not exist.
        """
        if self._timeout:
            get = retried(on=[NotFound], timeout=self._timeout)(self._ws.catalogs.get)
        else:
            get = self._ws.catalogs.get
        try:
            catalog_info = get(catalog.name)
            return Catalog(catalog_info.name)
        except (NotFound, TimeoutError):
            return None
        except DatabricksError as e:
            logger.warning(f"Unexpected error when getting catalog: {catalog.name}", exc_info=e)
            return None

    def _create_catalog(
        self,
        catalog: Catalog,
        catalog_storage: str,
        *,
        properties: dict[str, str] | None,
    ) -> Catalog:
        logger.info(f"Creating UC catalog: {catalog.name}")
        if catalog_storage == "metastore":
            self._ws.catalogs.create(catalog.name, comment="Created by UCX", properties=properties)
        else:
            self._ws.catalogs.create(
                catalog.name,
                storage_root=catalog_storage,
                comment="Created by UCX",
                properties=properties,
            )
        catalog_created = self._get_catalog(catalog)
        if catalog_created is None:
            raise NotFound(f"Created catalog '{catalog.name}' does not exist.")
        return catalog_created

    def _get_schema(self, schema: Schema) -> Schema | None:
        """Get a schema.

        Args:
            schema (Schema) : The schema to get.

        Returns:
            Schema : The schema it got.
            None : If the catalog does not exist.
        """
        if self._timeout:
            get = retried(on=[NotFound], timeout=self._timeout)(self._ws.schemas.get)
        else:
            get = self._ws.schemas.get
        try:
            schema_info = get(schema.full_name)
            return Schema(schema_info.catalog_name, schema_info.name)
        except (NotFound, TimeoutError):
            return None
        except DatabricksError as e:
            logger.warning(f"Unexpected error when getting schema: {schema.full_name}", exc_info=e)
            return None

    def _create_schema(self, schema: Schema) -> Schema:
        schema_existing = self._get_schema(schema)
        if schema_existing:
            logger.warning(f"Skipping already existing schema: {schema.full_name}")
            return schema_existing
        logger.info(f"Creating UC schema: {schema.full_name}")
        self._ws.schemas.create(schema.name, schema.catalog, comment="Created by UCX")
        schema_created = self._get_schema(schema)
        if schema_created is None:
            raise NotFound(f"Created schema '{schema.full_name}' does not exist.")
        return schema_created



import logging
from collections.abc import Iterable, Callable
from functools import cached_property

from databricks.labs.ucx.framework.owners import (
    Ownership,
    AdministratorLocator,
    LegacyQueryOwnership,
    WorkspacePathOwnership,
)
from databricks.labs.ucx.hive_metastore import TablesCrawler
from databricks.labs.ucx.hive_metastore.grants import GrantsCrawler, Grant
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatus
from databricks.labs.ucx.hive_metastore.tables import Table
from databricks.labs.ucx.source_code.base import UsedTable
from databricks.labs.ucx.source_code.used_table import UsedTablesCrawler
from databricks.labs.ucx.workspace_access.groups import GroupManager

logger = logging.getLogger(__name__)


class TableOwnership(Ownership[Table]):
    """Determine ownership of tables in the inventory based on the following rules:
    - If a table is owned by a principal in the grants table, then that principal is the owner.
    - If a table is written to by a query, then the owner of that query is the owner of the table.
    - If a table is written to by a notebook or file, then the owner of the path is the owner of the table.
    """

    def __init__(
        self,
        administrator_locator: AdministratorLocator,
        grants_crawler: GrantsCrawler,
        used_tables_in_paths: UsedTablesCrawler,
        used_tables_in_queries: UsedTablesCrawler,
        legacy_query_ownership: LegacyQueryOwnership,
        workspace_path_ownership: WorkspacePathOwnership,
    ) -> None:
        super().__init__(administrator_locator)
        self._grants_crawler = grants_crawler
        self._used_tables_in_paths = used_tables_in_paths
        self._used_tables_in_queries = used_tables_in_queries
        self._legacy_query_ownership = legacy_query_ownership
        self._workspace_path_ownership = workspace_path_ownership

    def _maybe_direct_owner(self, record: Table) -> str | None:
        owner = self._maybe_from_grants(record)
        if owner:
            return owner
        return self._maybe_from_sources(record)

    def _maybe_from_sources(self, record: Table) -> str | None:
        used_table = self._used_tables_snapshot.get((record.catalog, record.database, record.name))
        if not used_table:
            return None
        # If something writes to a table, then it's an owner of it
        if not used_table.is_write:
            return None
        if used_table.source_type == 'QUERY' and used_table.query_id:
            return self._legacy_query_ownership.owner_of(used_table.query_id)
        if used_table.source_type in {'NOTEBOOK', 'FILE'}:
            return self._workspace_path_ownership.owner_of_path(used_table.source_id)
        logger.warning(f"Unknown source type {used_table.source_type} for {used_table.source_id}")
        return None

    @cached_property
    def _used_tables_snapshot(self) -> dict[tuple[str, str, str], UsedTable]:
        index = {}
        for collection in (self._used_tables_in_paths.snapshot(), self._used_tables_in_queries.snapshot()):
            for used_table in collection:
                key = used_table.catalog_name, used_table.schema_name, used_table.table_name
                index[key] = used_table
        return index

    def _maybe_from_grants(self, record: Table) -> str | None:
        for grant in self._grants_snapshot:
            if not grant.action_type == 'OWN':
                continue
            object_type, full_name = grant.this_type_and_key()
            if object_type == 'TABLE' and full_name == record.key:
                return grant.principal
            if object_type in {'DATABASE', 'SCHEMA'} and full_name == f"{record.catalog}.{record.database}":
                return grant.principal
        return None

    @cached_property
    def _grants_snapshot(self):
        return self._grants_crawler.snapshot()


class DefaultSecurableOwnership(Ownership[Table]):
    """Determine ownership of tables in the inventory based on the following rules:
    -- If a global owner group is defined, then all tables are owned by that group.
    -- If the user running the application can be identified, then all tables are owned by that user.
    -- Otherwise, the table is owned by the administrator.
    """

    def __init__(
        self,
        administrator_locator: AdministratorLocator,
        table_crawler: TablesCrawler,
        group_manager: GroupManager,
        default_owner_group: str | None,
        app_principal_resolver: Callable[[], str | None],
    ) -> None:
        super().__init__(administrator_locator)
        self._tables_crawler = table_crawler
        self._group_manager = group_manager
        self._default_owner_group = default_owner_group
        self._app_principal_resolver = app_principal_resolver

    @cached_property
    def _application_principal(self) -> str | None:
        return self._app_principal_resolver()

    @cached_property
    def _static_owner(self) -> str | None:
        # If the default owner group is not valid, fall back to the application principal
        if self._default_owner_group and self._group_manager.current_user_in_owner_group(self._default_owner_group):
            logger.warning("Default owner group is not valid, falling back to administrator ownership.")
            return self._default_owner_group
        return self._application_principal

    def load(self) -> Iterable[Grant]:
        databases = set()
        catalogs = set()
        owner = self._static_owner
        if not owner:
            logger.warning("No owner found for tables and databases")
            return
        for table in self._tables_crawler.snapshot():
            table_name, view_name = self._names(table)

            if table.database not in databases:
                databases.add(table.database)
            if table.catalog not in catalogs:
                catalogs.add(table.catalog)
            yield Grant(
                principal=owner,
                action_type='OWN',
                catalog=table.catalog,
                database=table.database,
                table=table_name,
                view=view_name,
            )
        for database in databases:
            yield Grant(
                principal=owner,
                action_type='OWN',
                catalog="hive_metastore",
                database=database,
                table=None,
                view=None,
            )

        for catalog in catalogs:
            yield Grant(
                principal=owner,
                action_type='OWN',
                catalog=catalog,
                database=None,
                table=None,
                view=None,
            )

    @staticmethod
    def _names(table: Table) -> tuple[str | None, str | None]:
        if table.view_text:
            return None, table.name
        return table.name, None

    def _maybe_direct_owner(self, record: Table) -> str | None:
        return self._static_owner


class TableOwnershipGrantLoader:
    def __init__(self, tables_crawler: TablesCrawler, table_ownership: Ownership[Table]) -> None:
        self._tables_crawler = tables_crawler
        self._table_ownership = table_ownership

    def load(self) -> Iterable[Grant]:
        for table in self._tables_crawler.snapshot():
            owner = self._table_ownership.owner_of(table)
            table_name, view_name = self._names(table)
            yield Grant(
                principal=owner,
                action_type='OWN',
                catalog=table.catalog,
                database=table.database,
                table=table_name,
                view=view_name,
            )

    @staticmethod
    def _names(table: Table) -> tuple[str | None, str | None]:
        if table.view_text:
            return None, table.name
        return table.name, None


class TableMigrationOwnership(Ownership[TableMigrationStatus]):
    """Determine ownership of table migration records in the inventory.

    This is the owner of the source table, if (and only if) the source table is present in the inventory.
    """

    def __init__(self, tables_crawler: TablesCrawler, table_ownership: TableOwnership) -> None:
        super().__init__(table_ownership._administrator_locator)  # TODO: Fix this
        self._tables_crawler = tables_crawler
        self._table_ownership = table_ownership
        self._indexed_tables: dict[tuple[str, str], Table] | None = None

    def _tables_snapshot_index(self, reindex: bool = False) -> dict[tuple[str, str], Table]:
        index = self._indexed_tables
        if index is None or reindex:
            snapshot = self._tables_crawler.snapshot()
            index = {(table.database, table.name): table for table in snapshot}
            self._indexed_tables = index
        return index

    def _maybe_direct_owner(self, record: TableMigrationStatus) -> str | None:
        index = self._tables_snapshot_index()
        source_table = index.get((record.src_schema, record.src_table), None)
        return self._table_ownership.owner_of(source_table) if source_table is not None else None



import logging
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import replace

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatus, TableMigrationIndex
from databricks.labs.ucx.hive_metastore.tables import Table
from databricks.labs.ucx.hive_metastore.ownership import TableOwnership
from databricks.labs.ucx.progress.history import ProgressEncoder
from databricks.labs.ucx.progress.install import Historical
from databricks.labs.ucx.source_code.base import UsedTable
from databricks.labs.ucx.source_code.used_table import UsedTablesCrawler


logger = logging.getLogger(__name__)


class TableProgressEncoder(ProgressEncoder[Table]):
    """Encoder class:Table to class:History."""

    def __init__(
        self,
        sql_backend: SqlBackend,
        ownership: TableOwnership,
        migration_status_refresher: CrawlerBase[TableMigrationStatus],
        used_tables_crawlers: list[UsedTablesCrawler],
        run_id: int,
        workspace_id: int,
        catalog: str,
    ) -> None:
        super().__init__(
            sql_backend,
            ownership,
            Table,
            run_id,
            workspace_id,
            catalog,
            "multiworkspace",
            "historical",
        )
        self._migration_status_refresher = migration_status_refresher
        self._used_tables_crawlers = used_tables_crawlers

    def append_inventory_snapshot(self, snapshot: Iterable[Table]) -> None:
        migration_index = TableMigrationIndex(self._migration_status_refresher.snapshot())
        used_hive_tables = self._get_used_hive_tables()
        history_records = []
        for record in snapshot:
            history_record = self._encode_table_as_historical(record, migration_index, used_hive_tables)
            history_records.append(history_record)
        logger.debug(f"Appending {len(history_records)} {self._klass} table record(s) to history.")
        # The mode is 'append'. This is documented as conflict-free.
        self._sql_backend.save_table(escape_sql_identifier(self.full_name), history_records, Historical, mode="append")

    def _get_used_hive_tables(self) -> dict[str, list[UsedTable]]:
        used_tables: dict[str, list[UsedTable]] = defaultdict(list[UsedTable])
        for crawler in self._used_tables_crawlers:
            for used_table in crawler.snapshot():
                if used_table.catalog_name == "hive_metastore":
                    used_tables[used_table.full_name].append(used_table)
        return used_tables

    def _encode_table_as_historical(
        self, record: Table, migration_index: TableMigrationIndex, used_hive_tables: dict[str, list[UsedTable]]
    ) -> Historical:
        """Encode a table record, enriching with the migration status and used table references.

        Possible failures, the table is
        - Pending migration
        - A Hive table referenced by code

        Grants are purposefully left out, because a grant might not be mappable to UC, like `READ_METADATA`, thus
        possibly resulting in false "pending migration" failure for tables that are migrated to UC with their relevant
        grants also being migrated.
        """
        historical = super()._encode_record_as_historical(record)
        failures = []
        if not migration_index.is_migrated(record.database, record.name):
            failures.append("Pending migration")
        for used_table in used_hive_tables.get(record.full_name, []):
            failures.append(f"Used by {used_table.source_type}: {used_table.source_id}")
        return replace(historical, failures=historical.failures + failures)



from dataclasses import replace

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.hive_metastore.grants import Grant, GrantOwnership
from databricks.labs.ucx.progress.history import ProgressEncoder
from databricks.labs.ucx.progress.install import Historical


class GrantProgressEncoder(ProgressEncoder[Grant]):
    """Encoder class:Grant to class:History.

    A failure for a grants implies it cannot be mapped to Unity Catalog.
    """

    def __init__(
        self,
        sql_backend: SqlBackend,
        ownership: GrantOwnership,
        run_id: int,
        workspace_id: int,
        catalog: str,
        schema: str = "multiworkspace",
        table: str = "historical",
    ) -> None:
        super().__init__(sql_backend, ownership, Grant, run_id, workspace_id, catalog, schema, table)

    def _encode_record_as_historical(self, record: Grant) -> Historical:
        historical = super()._encode_record_as_historical(record)
        failures = []
        if not record.uc_grant_sql():
            type_, key = record.this_type_and_key()
            failures = [f"Action '{record.action_type}' on {type_} '{key}' unmappable to Unity Catalog"]
        return replace(historical, failures=historical.failures + failures)
